@@ -7,24 +7,29 @@
  *  By Matti Aarnio <mea@utu.fi> on 26-Sep-94
  */
 
-#include <stdio.h>
+#include "mailer.h"
+#include "libz.h"
 
 int cfgets(s, n, stream)
 	char *s;
 	int n;
 	FILE *stream;
 {
-	int cnt = 0;
-	int c;
+	register int cnt = 0;
 
-	while (n > 0 && (c = getc(stream)) != EOF) {
-	  *s++ = c;
-	  ++cnt;
-	  --n;
-	  if (c == '\n')
+	while (n > 0) {
+	  register int c = getc(stream);
+	  if (c != EOF) {
+	    *s++ = c;
+	    ++cnt;
+	    --n;
+	    if (c == '\n')
+	      break;
+	  } else {
+	    if (cnt == 0)
+	      return EOF;
 	    break;
+	  }
 	}
-	if (cnt == 0 && c == EOF)
-	  return EOF;
 	return cnt;
 }
