@@ -189,7 +189,7 @@ mail_open(type)
 	cp = (host == NULL) ? "I" : host ;
 	scratch = mail_alloc(strlen(PUBLICDIR)+strlen(cp)+3+1+10);
 
-	sprintf(scratch, "%s/%7s:%d%%", PUBLICDIR, cp, getpid());
+	sprintf(scratch, "%s/%7s:%d%%", PUBLICDIR, cp, (int)getpid());
 
 	fp = _mail_fopen(&scratch);
 	if (fp == NULL) {
@@ -279,11 +279,29 @@ mail_open(type)
 	      cp = getenv("USER");
 	    if (cp == NULL)
 	      cp = "\"??\"";
-	    fprintf(fp, "rcvdfrom %s@%s\n", cp, namebuf);
+	    fprintf(fp, "rcvdfrom STDIN (%s@%s)\n", cp, namebuf);
 	  }
 	return fp;
 }
 
+
+/*
+ * Return currently open spool file name
+ */
+
+char *
+mail_fname(fp)
+	FILE *fp;
+{
+	int fd = FILENO(fp);
+
+	if (fd < 0 || fd >= mail_nfiles)
+	  return NULL;
+
+	return mail_file[fd];
+}
+
+
 /*
  * Abort the message file composition on the indicated stream.
  */
