@@ -9,26 +9,29 @@
  */
 #include <stdio.h>
 #include <sys/types.h>
-#include "malloc.h"
 
-char *
+#include "hostenv.h"
+#include "zmalloc.h"
+#include "libc.h"
+
+void *
 mail_alloc(nbytes)
 	unsigned int nbytes;
 {
-	return (char*)malloc(nbytes);
+	return (void*)malloc(nbytes);
 }
 
-char *
+void *
 mail_realloc(oldptr,nbytes)
 	unsigned int nbytes;
-	char *oldptr;
+	void *oldptr;
 {
-	return (char*)realloc(oldptr,nbytes);
+	return (void*)realloc(oldptr,nbytes);
 }
 
 void
 mail_free(s)
-	char *s;
+	void *s;
 {
 	free(s);
 }
