@@ -1,6 +1,10 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *
+ *	Modifications/maintance, Matti Aarnio, over years 1990-1998
+ *
+ *	'longestmatch' driver kissg@sztaki.hu 970209
  */
 
 /*
@@ -54,7 +58,7 @@ struct db_info {
 	time_t		ttl;			/* default time to live */
 	conscell	*DBFUNCD(driver);	/* completion routine */
 	conscell	*DBFUNC(lookup);	/* low-level lookup routine */
-	void		DBFUNC(close);		/* flush buffered data, close*/
+	void		DBFUNCV(close);		/* flush buffered data, close*/
 	int		DBFUNCV(add);		/* add key/value pair */
 	int		DBFUNC(remove);		/* remove key */
 	void		DBFUNCF(print);		/* print database */
@@ -82,80 +86,123 @@ struct db_kind {
 		  count_core,
 		  owner_core,
 		  NULL,
-		  Nul } },
+		  Nul,
+		  NULL } },
 { "header",	{ NULL, NULL, 0, 0, 0, NULL, search_header, close_header,
 		  add_header, remove_header, print_header, count_header,
-		  owner_header, NULL, Nul } },
+		  owner_header, NULL, Nul, NULL } },
 { "unordered",	{ NULL, NULL, 0, 10, 0, NULL, search_seq, close_seq,
 		  add_seq, NULL, print_seq, count_seq, owner_seq, modp_seq,
-		  Nul } },
+		  Nul, NULL } },
 #ifndef	HAVE_MMAP
 { "ordered",	{ NULL, NULL, 0, 10, 0, NULL, search_bin, close_seq,
 		  NULL, NULL, print_seq, count_seq, owner_seq, modp_seq,
-		  Nul } },
+		  Nul, NULL } },
 #else /* HAVE_MMAP */ /* When using MMAP(), no cache is needed for ordered.. */
 { "ordered",	{ NULL, NULL, 0, 0, 0, NULL, search_bin, close_seq,
 		  NULL, NULL, print_seq, count_seq, owner_seq, modp_seq,
-		  Nul } },
+		  Nul, NULL } },
 #endif
 
 #ifdef	HAVE_RESOLVER
-{ "hostsfile",	{ "/etc/hosts", NULL, 0, 10, 0, NULL, search_hosts, NULL,
-		  NULL, NULL, print_hosts, NULL, NULL, NULL, Nul } },
+{ "hostsfile",	{ "/etc/hosts", NULL, 0, 0, 0, NULL, search_hosts, NULL,
+		  NULL, NULL, print_hosts, NULL, NULL, NULL, Nul, NULL } },
 #endif	/* HAVE_RESOLVER */
 #ifdef	HAVE_RESOLVER
 #ifndef RESOLV_CONF
 # define RESOLV_CONF "/etc/resolv.conf"
 #endif
-{ "bind",	{ RESOLV_CONF, NULL, 0, 10, 0, NULL,
-		  search_res, NULL, NULL, NULL, NULL, NULL, NULL, NULL, Nul }},
+{ "bind",	{ RESOLV_CONF, NULL, 0, 0, 0, NULL, search_res, NULL, NULL,
+		    NULL, NULL, NULL, NULL, NULL, Nul, NULL }},
 #endif	/* HAVE_RESOLV */
-{ "selfmatch",	{ NULL, NULL, 0, 0, 0, NULL, search_selfmatch, NULL, NULL,
-		  NULL, print_selfmatch, count_selfmatch, NULL, NULL, Nul } },
+{ "selfmatch",	{ NULL, NULL, 0, 0, 0, NULL, search_selfmatch, NULL, NULL,NULL,
+		  print_selfmatch, count_selfmatch, NULL, NULL, Nul, NULL } },
 #ifdef	HAVE_NDBM_H
-{ "ndbm",	{ NULL, NULL, 0, 10, 0, NULL, search_ndbm, close_ndbm,
+{ "ndbm",	{ NULL, NULL, 0, 0, 0, NULL, search_ndbm, close_ndbm,
 		  add_ndbm, remove_ndbm, print_ndbm, count_ndbm, owner_ndbm,
-		  modp_ndbm, Nul } },
+		  modp_ndbm, Nul, NULL } },
 #endif	/* HAVE_NDBM */
 #ifdef	HAVE_GDBM_H
-{ "gdbm",	{ NULL, NULL, 0, 10, 0, NULL, search_gdbm, close_gdbm,
+{ "gdbm",	{ NULL, NULL, 0, 0, 0, NULL, search_gdbm, close_gdbm,
 		  add_gdbm, remove_gdbm, print_gdbm, count_gdbm, owner_gdbm,
-		  modp_gdbm, Nul } },
+		  modp_gdbm, Nul, NULL } },
 #endif	/* HAVE_GDBM */
 #ifdef	HAVE_DBM
-{ "dbm",	{ NULL, NULL, 0, 10, 0, NULL, search_dbm, close_dbm,
+{ "dbm",	{ NULL, NULL, 0, 0, 0, NULL, search_dbm, close_dbm,
 		  add_dbm, remove_dbm, print_dbm, count_dbm, owner_dbm,
-		  NULL, Nul } },
+		  NULL, Nul, NULL } },
 #endif	/* HAVE_DBM */
 #ifdef	HAVE_DB_H
-{ "btree",	{ NULL, NULL, 0, 10, 0, NULL, search_btree, close_btree,
+{ "btree",	{ NULL, NULL, 0, 0, 0, NULL, search_btree, close_btree,
 		  add_btree, remove_btree, print_btree, count_btree,
-		  owner_btree, modp_btree, Nul } },
-{ "bhash",	{ NULL, NULL, 0, 10, 0, NULL, search_bhash, close_bhash,
+		  owner_btree, modp_btree, Nul, NULL } },
+{ "bhash",	{ NULL, NULL, 0, 0, 0, NULL, search_bhash, close_bhash,
 		  add_bhash, remove_bhash, print_bhash, count_bhash,
-		  owner_bhash, modp_bhash, Nul } },
+		  owner_bhash, modp_bhash, Nul, NULL } },
 #endif	/* HAVE_DB_H */
 #ifdef	HAVE_YP
-{ "yp",		{ NULL, NULL, 0, 10, 0, NULL, search_yp, NULL,
-		  NULL, NULL, print_yp, NULL, owner_yp, NULL, Nul } },
+{ "yp",		{ NULL, NULL, 0, 0, 0, NULL, search_yp, NULL, NULL,
+		  NULL, print_yp, NULL, owner_yp, NULL, Nul, NULL } },
 #endif	/* HAVE_YP */
 #ifdef HAVE_LDAP
-{ "ldap",	{ NULL, NULL, 0, 10, 0, NULL, search_ldap, close_ldap,
-		  NULL, NULL, NULL, NULL, NULL, modp_ldap, Nul } },
+{ "ldap",	{ NULL, NULL, 0, 0, 0, NULL, search_ldap, close_ldap,
+		  NULL, NULL, NULL, NULL, NULL, modp_ldap, Nul, NULL } },
 #endif
-{ 0 }	/* proto_config is initialized from this entry */
+{ NULL, { NULL, NULL, 0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+	    /* proto_config is initialized from this entry */
+	    NULL, NULL, Nul, NULL }}
 };
 
 
 /* drivers */
 static conscell	*find_domain       __((conscell *DBFUNC(lookupfn), search_info *sip));
 static conscell	*find_nodot_domain __((conscell *DBFUNC(lookupfn), search_info *sip));
+static conscell *find_longest_match __((conscell *DBFUNC(lookupfn), search_info *sip));
 /* others.. */
 static void      cacheflush __((struct db_info *dbip));
 extern conscell	*readchunk  __((const char *file, long offset));
 static int	 iclistdbs  __((struct spblk *spl));
 
 
+#define	CACHE(x,f)	((dbip->cache+(x))->f)
+
+static void (*cachemarkupfunc) __((conscell*));
+static int
+iccachemarkup(spl)
+	struct spblk *spl;
+{
+	struct db_info *dbip = (struct db_info *)spl->data;
+	int i;
+	conscell *ll;
+
+	if (dbip == NULL || dbip->cache_size == 0 || dbip->cache == NULL)
+		return 0;
+
+	/* flush cache */
+	for (i = 0; i < dbip->cache_size && CACHE(i,key) != NULL; ++i) {
+		ll = CACHE(i,value);
+		cachemarkupfunc(ll);
+	}
+	return 0;
+}
+
+static void
+cache_gc_markup_iterator(mrkupfunc)
+     void (*mrkupfunc)__((conscell*));
+{
+  cachemarkupfunc = mrkupfunc;
+  sp_scan(iccachemarkup, (struct spblk *)NULL, spt_databases);
+}
+
+static void
+register_cache_gc_markup_iterator __((void))
+{
+  static int done = 0;
+  if (done) return;
+  functionprot(cache_gc_markup_iterator);
+  done = 1;
+}
+
 
 /*
  * Define a new relation according to the command line arguments.
@@ -170,19 +217,23 @@ run_relation(argc, argv)
 	struct db_info proto_config, *dbip;
 	struct db_kind *dbkp;
 	struct spblk *spl;
-	int c, errflg, set_cache_size;
+	int c, errflg, set_cache_size, dbtest;
 	spkey_t symid;
 	memtypes oval;
 	char *cp, *dbtyp;
 
+	if (spt_files == NULL)          spt_files          = sp_init();
+	if (spt_files->symbols == NULL) spt_files->symbols = sp_init();
+
 	errflg = 0;
 	dbtyp = NULL;
 	set_cache_size = 0;
 	optind = 1;
+	dbtest = 0;
 	proto_config
 		= db_kinds[sizeof db_kinds/(sizeof (struct db_kind))-1].config;
 	while (1) {
-		c = getopt(argc, (char*const*)argv, "Cbilmnpud:f:s:t:e:");
+		c = getopt(argc, (char*const*)argv, "Cbilmnpud:f:s:t:Te:");
 		if (c == EOF)
 			break;
 		switch (c) {
@@ -194,6 +245,8 @@ run_relation(argc, argv)
 				proto_config.driver = find_nodot_domain;
 			else if (strcmp(optarg, "pathalias") == 0)
 				proto_config.driver = find_domain;
+			else if (strcmp(optarg, "longestmatch") == 0)
+				proto_config.driver = find_longest_match;
 			else
 				++errflg;
 			break;
@@ -234,6 +287,9 @@ run_relation(argc, argv)
 			}
 			proto_config.subtype = cp;
 			break;
+		case 'T':
+			dbtest = 1;
+			break;
 		case 'e':	/* expiry - cache data time to live */
 			proto_config.ttl = atol(optarg);
 			break;
@@ -251,16 +307,19 @@ run_relation(argc, argv)
 		fprintf(stderr,
 		"Usage: %s -t dbtype[,subtype] [-f file -e# -s# -bilmnpu -d driver] name\n",
 			argv[0]);
+		fprintf(stderr,
+		       "       %s -T -t dbtype dummyname\n", argv[0]);
 		fprintf(stderr,
 			"      dbtypes: ");
 		for (dbkp = &db_kinds[0];
 		     dbkp < &db_kinds[(sizeof db_kinds)/sizeof (struct db_kind)];
 		     ++dbkp) {
-		  if (dbkp->name != NULL)
+		  if (dbkp->name != NULL) {
 		    if (dbkp == &db_kinds[0])
 		      fprintf(stderr,"%s",dbkp->name);
 		    else
 		      fprintf(stderr,",%s",dbkp->name);
+		  }
 		}
 		fprintf(stderr,"\n");
 		return 1;
@@ -279,11 +338,14 @@ run_relation(argc, argv)
 	}
 	/* the -1 in the following compensates for the terminating null entry */
 	if (dbkp >= &db_kinds[((sizeof db_kinds)/sizeof (struct db_kind))-1]) {
-		fprintf(stderr,
-			"relation: I don't know about the %s database type!\n",
-			dbtyp);
-		return 3;
+	  if (!dbtest)
+	    fprintf(stderr,
+		    "relation: I don't know about the %s database type!\n",
+		    dbtyp);
+	  return 3;
 	}
+	if (dbtest)
+	  return 0;
 
 	symid = symbol(argv[optind]); /* Database name symbol  */
 	if (sp_lookup(symid, spt_databases) != NULL) {
@@ -361,6 +423,7 @@ run_relation(argc, argv)
 	} else
 		dbip->cache = NULL;	/* superfluous, but why not ... */
 	sp_install(symid, dbip, 0, spt_databases);
+	register_cache_gc_markup_iterator();
 	spl = sp_lookup(symbol(DBLOOKUPNAME), spt_builtins);
 	if (spl == NULL) {
 		fprintf(stderr, "%s: '%s' isn't built in\n",
@@ -481,7 +544,7 @@ run_db(argc, argv)
 	const char *argv[];
 {
 	int errflag;
-	struct db_info *dbip;
+	struct db_info *dbip = NULL;
 	search_info si;
 	struct spblk *spl;
 
@@ -579,7 +642,7 @@ run_db(argc, argv)
 		 * file/db name entry from the splay tree, because there may
 		 * be multple references to it.  For example: /dev/null.
 		 */
-		(*dbip->close)(&si);
+		(*dbip->close)(&si,"db flush");
 
 		/*
 		 * Close auxiliary (indirect) data file as well (aliases.dat).
@@ -588,7 +651,7 @@ run_db(argc, argv)
 		 */
 		if (dbip->postproc == Indirect) {
 			si.file = dbip->subtype;
-			(*dbip->close)(&si);
+			(*dbip->close)(&si,"db flush indirect");
 		}
 		break;
 	case 'o':	/* owner */
@@ -624,8 +687,6 @@ run_db(argc, argv)
 }
 
 
-#define	CACHE(x,f)	((dbip->cache+(x))->f)
-
 /*
  * This is the basic interface to the database lookup routines.
  * It uses the configuration information for each relation properly,
@@ -636,7 +697,7 @@ conscell *
 db(dbname, key)
 	const char *dbname, *key;
 {
-	register int i, j, k;
+	register int i, j, k, keylen;
 	memtypes oval = stickymem;
 	conscell *l, *ll, *tmp;
 	struct spblk *spl;
@@ -645,6 +706,11 @@ db(dbname, key)
 	search_info si;
 	struct cache tce;
 	char kbuf[BUFSIZ];	/* XX: */
+	int slen;
+	GCVARS3;
+
+	if (spt_files == NULL)          spt_files          = sp_init();
+	if (spt_files->symbols == NULL) spt_files->symbols = sp_init();
 
 	if (key == NULL || *key == '\0') {
 		fprintf(stderr,
@@ -662,16 +728,21 @@ db(dbname, key)
 		fprintf(stderr, "%s(%s)\n", dbname, key);
 	/* apply flags */
 	realkey = NULL;
+
 	if (dbip->flags & DB_MAPTOLOWER) {
-		strncpy(kbuf, key, sizeof(kbuf));
-		kbuf[sizeof(kbuf)-1] = 0;
-		strlower(kbuf);
-		key = kbuf;
+	  keylen = strlen(key)+1;
+	  if (keylen > sizeof(kbuf)) keylen = sizeof(kbuf);
+	  memcpy(kbuf, key, keylen); /* was: strncpy */
+	  kbuf[sizeof(kbuf)-1] = 0;
+	  strlower(kbuf);
+	  key = kbuf;
 	} else if (dbip->flags & DB_MAPTOUPPER) {
-		strncpy(kbuf, key, sizeof(kbuf));
-		kbuf[sizeof(kbuf)-1] = 0;
-		strupper(kbuf);
-		key = kbuf;
+	  keylen = strlen(key)+1;
+	  if (keylen > sizeof(kbuf)) keylen = sizeof(kbuf);
+	  memcpy(kbuf, key, keylen); /* was: strncpy */
+	  kbuf[sizeof(kbuf)-1] = 0;
+	  strupper(kbuf);
+	  key = kbuf;
 	}
 	si.file = dbip->file;
 	si.key  = key;
@@ -681,10 +752,10 @@ db(dbname, key)
 	    && dbip->modcheckp != NULL && (*dbip->modcheckp)(&si)) {
 		if (dbip->close != NULL) {
 			cacheflush(dbip);
-			(*dbip->close)(&si);
+			(*dbip->close)(&si,"db modcheck");
 			if (dbip->postproc == Indirect) {
 				si.file = dbip->subtype;
-				(*dbip->close)(&si);
+				(*dbip->close)(&si,"db modcheck indirect");
 				si.file = dbip->file;
 			}
 		}
@@ -699,10 +770,12 @@ db(dbname, key)
 					fprintf(stderr,
 						"... expiring %s from cache\n",
 						CACHE(i,key));
-				if ((ll = CACHE(i,value)) != NULL)
-					s_free_tree(ll);
+				/* if ((ll = CACHE(i,value)) != NULL)
+				   s_free_tree(ll); */
+				CACHE(i,value) = NULL;
 				if (CACHE(i,key) != NULL) /* always true */
 					free((char *)CACHE(i,key));
+				CACHE(i,key) = NULL;
 				continue;
 			}
 			if (i > j) { /* expiry */
@@ -716,7 +789,14 @@ db(dbname, key)
 				fprintf(stderr,
 					"... comparing %s and %s in cache\n",
 					CACHE(i,key), key);
-			if (CISTREQ(CACHE(i,key), key)) {
+
+			/* Match mixed case IF the mapping is to lower or
+			   to upper, if not, match as is! */
+			if (((dbip->flags & (DB_MAPTOUPPER|DB_MAPTOLOWER)) &&
+			     CISTREQ(CACHE(i,key), key)) ||
+			    (!(dbip->flags & (DB_MAPTOUPPER|DB_MAPTOLOWER)) &&
+			     strcmp(CACHE(i,key), key) == 0)) {
+
 				/* cache maintenance */
 				tce = *(dbip->cache+i);
 				if (j > 0 && D_db)
@@ -742,7 +822,8 @@ db(dbname, key)
 				if (D_db)
 					fprintf(stderr, "... found in cache\n");
 				/* return a scratch value */
-				return s_copy_tree(dbip->cache->value);
+				tmp = s_copy_chain(dbip->cache->value);
+				return tmp;
 			}
 			++j;
 		}
@@ -751,9 +832,17 @@ db(dbname, key)
 		oval = stickymem;	/* cannot return before this is reset */
 		stickymem = MEM_MALLOC;	/* this is reset down below */
 		/* key gets clobbered somewhere, so save it here */
-		realkey = strsave(key);
+		realkey = strdup(key);
 	}
-	l = NULL;
+
+	l = ll = tmp = NULL;
+	GCPRO3(l, ll, tmp);
+
+	if (D_db) {
+	  fprintf(stderr, "%s(%s) = ", dbname, key);
+	  fflush(stderr);
+	}
+
 	if ((dbip->driver == NULL &&
 	     (l = (*dbip->lookup)(&si)) != NULL) ||
 	    (dbip->driver != NULL &&
@@ -761,17 +850,20 @@ db(dbname, key)
 
 		switch (dbip->postproc) {
 		case Boolean:
-			if (stickymem == MEM_MALLOC)
-				s_free_tree(l);
-			l = newstring(strsave(key));
+			/* if (stickymem == MEM_MALLOC)
+			   s_free_tree(l); */
+			slen = strlen(key);
+			l = newstring(dupnstr(key, slen), slen);
 			break;
 		case NonNull:
 			if (STRING(l) && *(l->string) == '\0') {
-				if (stickymem == MEM_MALLOC)
-					s_free_tree(l);
-				l = newstring(strsave(key));
+				/* if (stickymem == MEM_MALLOC)
+				   s_free_tree(l); */
+				slen = strlen(key);
+				l = newstring(dupnstr(key, slen), slen);
 			} else if (LIST(l) && car(l) == NULL) {
-				car(l) = newstring(strsave(key));
+				slen = strlen(key);
+				car(l) = newstring(dupnstr(key, slen), slen);
 			}
 			break;
 		case Indirect:
@@ -780,14 +872,14 @@ db(dbname, key)
 			 * the file named by the subtype.  Used for aliases.
 			 */
 			if (LIST(l) || !isdigit(*(l->string))) {
-				if (stickymem == MEM_MALLOC)
-					s_free_tree(l);
+				/* if (stickymem == MEM_MALLOC)
+				   s_free_tree(l); */
 				l = NULL;
 				break;
 			}
-			i = atoi(l->string);	/* file offset */
-			if (stickymem == MEM_MALLOC)
-				s_free_tree(l);
+			i = atol(l->string);	/* file offset */
+			/* if (stickymem == MEM_MALLOC)
+			   s_free_tree(l); */
 			l = readchunk(dbip->subtype, (long)i);
 			break;
 		case Pathalias:
@@ -797,19 +889,22 @@ db(dbname, key)
 			break;
 		}
 		if (D_db) {
-			fprintf(stderr, "%s(%s) = ", dbname, key);
 			s_grind(l, stderr);
 			putc('\n', stderr);
 		}
 	} else if (dbip->postproc == NonNull) {
 		if (D_db)
-			fprintf(stderr, "%s(%s) = %s\n", dbname, key, key);
-		l = newstring(strsave(key));
+			fprintf(stderr, "%s\n", key);
+		slen = strlen(key);
+		l = newstring(dupnstr(key, slen), slen);
 	} else {
+		if (D_db)
+			fprintf(stderr, "NIL\n");
 		if (dbip->cache_size > 0) {
 			stickymem = oval;
 			free(realkey);
 		}
+		UNGCPRO3;
 		return NULL;
 	}
 	if (!deferit && dbip->cache_size > 0) {
@@ -820,8 +915,8 @@ db(dbname, key)
 			if (D_db)
 				fprintf(stderr, "... freeing cache[%d]: %s\n",
 						i, CACHE(i,key));
-			if ((ll = CACHE(i,value)) != NULL)
-				s_free_tree(ll);
+			/* if ((ll = CACHE(i,value)) != NULL)
+			   s_free_tree(ll); */
 			free((char *)CACHE(i,key));
 		} else if (i < dbip->cache_size - 1) {
 			if (D_db)
@@ -847,15 +942,18 @@ db(dbname, key)
 				fprintf(stderr, "\n");
 		}
 		stickymem = oval;
-		ll = s_copy_tree(l);	/* scratch version returned to shell */
+		ll = l;
+
 	} else if (dbip->cache_size > 0) {
+
 		stickymem = oval;
-		ll = s_copy_tree(l);	/* scratch version returned to shell */
 		free(realkey);
-		s_free_tree(l);
+		ll = l;
+
 	} else
 		ll = l;
 
+	UNGCPRO3;
 	return ll;
 }
 
@@ -884,22 +982,22 @@ cacheflush(dbip)
 	struct db_info *dbip;
 {
 	int i;
-	conscell *ll;
 
 	if (dbip == NULL || dbip->cache_size == 0 || dbip->cache == NULL)
 		return;
 
 	/* flush cache */
 	for (i = 0; i < dbip->cache_size && CACHE(i,key) != NULL; ++i) {
-		ll = CACHE(i,value);
-		if (ll != NULL)
-			s_free_tree(ll);
+		/* conscell *ll = CACHE(i,value);
+		   if (ll != NULL)
+		     s_free_tree(ll); */
 		if (CACHE(i,key) != NULL) /* always true */
-			free((char *)CACHE(i,key));
+		  free((char *)CACHE(i,key));
 	}
-	dbip->cache->key = NULL;	/* ensure terminator */
+	CACHE(0,key) = NULL;	/* ensure terminator */
 }
 
+
 #ifdef	MALLOC_TRACE
 
 /*
@@ -1082,3 +1180,89 @@ find_nodot_domain(lookupfn, sip)
 #endif
 	return l;
 }
+
+
+/*
+ * Searching the longest match.
+ *
+ * The lookup sequence for foo.bar.edu is:
+ *
+ *	foo.bar.edu
+ *	.bar.edu
+ *	.edu
+ *	.
+ *
+ * The lookup sequence for 1.2.3.13 is:
+ *
+ *	1.2.3.13/32
+ *	1.2.3.12/31
+ *	1.2.3.12/30
+ *	1.2.3.8/29
+ *	1.2.3.0/28
+ *      ...
+ *	1.0.0.0/8
+ *	...
+ *	0.0.0.0/1
+ *	0.0.0.0/0
+ */
+
+static conscell *
+find_longest_match(lookupfn, sip)
+	conscell *DBFUNC(lookupfn);
+	search_info *sip;
+{
+	register char *cp;
+	conscell *l;
+	char buf[BUFSIZ]; 
+	char *realkey;
+	unsigned int oct1,oct2,oct3,oct4;
+
+	if (sscanf((char*)(sip->key[0]=='[' ? sip->key+1 : sip->key),
+	    "%3u.%3u.%3u.%3u",
+	    &oct1,&oct2,&oct3,&oct4) == 4) { /* IP address with optional [] */
+		unsigned int h_addr,h_mask;
+		int prefix;
+
+		h_addr = (((oct1 & 255) << 24) |
+			  ((oct2 & 255) << 16) |
+			  ((oct3 & 255) <<  8) |
+			  ((oct4 & 255)));
+
+		sip->key=buf;
+		for (prefix=32, h_mask=0xffffffffL;
+		    prefix>=0;  --prefix, h_mask<<=1) {
+			sprintf((char*)buf,"%u.%u.%u.%u/%d",
+				((h_addr&h_mask) >> 24) & 255,
+				((h_addr&h_mask) >> 16) & 255,
+				((h_addr&h_mask) >>  8) & 255,
+				((h_addr&h_mask)      ) & 255,
+				prefix);
+			if ((l = (*lookupfn)(sip)) != NULL)
+				return l;
+		}
+	}
+	else {	/* domain name */
+		/* check the key as given */
+		if ((l = (*lookupfn)(sip)) != NULL)
+			return l;
+		realkey = (char *) sip->key;
+		/* iterate over the superdomains of the key */
+		for (cp = realkey; *cp;) {
+			while (*cp && *cp != '.')
+				++cp;
+			while (*cp == '.')
+				++cp;
+			if (*(cp-1) == '.') {
+				sip->key = cp-1;
+				if ((l = (*lookupfn)(sip)) != NULL)
+					return l;
+			}
+		}
+		/* Still failed ?  Try to look for "." */
+		sip->key = ".";
+		if ((l = (*lookupfn)(sip)) != NULL)
+			return l;
+	}
+	
+	return NULL;
+}
