@@ -1,7 +1,7 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Some functions Copyright 1991-1997 Matti Aarnio.
+ *	Some functions Copyright 1991-1998 Matti Aarnio.
  */
 
 /*
@@ -38,6 +38,7 @@
 #endif /* HAVE_DIRENT_H */
 
 #include "zmsignal.h"
+#include "zsyslog.h"
 #include "mail.h"
 #include "interpret.h"
 #include "io.h"
@@ -65,8 +66,12 @@ extern conscell *sh_car       __((conscell *avl, conscell *il));
 static conscell *run_cadr     __((conscell *avl, conscell *il));
 static conscell *run_caddr    __((conscell *avl, conscell *il));
 static conscell *run_cadddr   __((conscell *avl, conscell *il));
-static conscell *run_listexpand __((conscell *avl, conscell *il));
+static conscell *run_listexpand   __((conscell *avl, conscell *il));
+#if 0
+static conscell *run_newattribute __((conscell *avl, conscell *il));
+#endif
 
+static int run_doit      ARGCV;
 static int run_stability ARGCV;
 static int run_process   ARGCV;
 static int run_grpmems   ARGCV;
@@ -86,6 +91,7 @@ static int run_squirrel  ARGCV;
 static int run_822syntax ARGCV;
 static int run_dequote   ARGCV;
 static int run_condquote ARGCV;
+static int run_syslog    ARGCV;
 
 #if	defined(XMEM) && defined(CSRIMALLOC)
 static int run_malcontents ARGCV;
@@ -96,6 +102,7 @@ extern int nobody;
 extern struct group  *getgrnam __((const char *));
 extern struct passwd *getpwnam __((const char *));
 extern time_t time __((time_t *));
+extern int routerdirloops;
 
 #ifndef strchr
 extern char *strchr(), *strrchr();
@@ -117,6 +124,7 @@ struct shCmd fnctns[] = {
 {	"user",		NULL,	run_caddr,	NULL,	SH_ARGV	},
 {	"attributes",	NULL,	run_cadddr,	NULL,	SH_ARGV	},
 {	"stability",	run_stability,	NULL,	NULL,	0	},
+{	"stableprocess", run_doit,	NULL,	NULL,	0	},
 {	"daemon",	run_daemon,	NULL,	NULL,	0	},
 {	"process",	run_process,	NULL,	NULL,	0	},
 {	"rfc822",	run_rfc822,	NULL,	NULL,	0	},
@@ -124,6 +132,9 @@ struct shCmd fnctns[] = {
 {	"printaliases",	run_praliases,	NULL,	NULL,	0	},
 {	"listaddresses",run_listaddrs,	NULL,	NULL,	SH_ARGV	},
 {	"listexpand",	NULL,	run_listexpand,	NULL,	SH_ARGV	},
+#if 0
+{	"newattribute",	NULL,	run_newattribute, NULL,	SH_ARGV	},
+#endif
 {	"homedirectory",run_homedir,	NULL,	NULL,	0	},
 {	"rfc822date",	run_822date,	NULL,	NULL,	0	},
 {	"filepriv",	run_filepriv,	NULL,	NULL,	SH_ARGV	},
@@ -137,11 +148,13 @@ struct shCmd fnctns[] = {
 {	"rfc822syntax",	run_822syntax,	NULL,	NULL,	0	},
 {	"dequote",	run_dequote,	NULL,	NULL,	0	},
 {	"condquote",	run_condquote,	NULL,	NULL,	0	},
+{	"syslog",	run_syslog,	NULL,	NULL,	0	},
+{	"logger",	run_syslog,	NULL,	NULL,	0	},
 #if	defined(XMEM) && defined(CSRIMALLOC)
 {	"malcontents",	run_malcontents,NULL,	NULL,	0	},
 #endif	/* CSRIMALLOC */
 /* The rest have been added locally */
-{ 0 }
+{ NULL, NULL, NULL, NULL, 0 }
 };
 
 int		funclevel = 0;
@@ -230,6 +243,33 @@ run_trace(argc, argv)
 	return 0;
 }
 
+
+int gensym;
+const char *gs_name = "g%d";
+
+static int
+run_gensym(argc, argv)
+	int argc;
+	const char *argv[];
+{
+	printf(gs_name, gensym++);
+	putchar('\n');
+	return 0;
+}
+
+static void
+free_gensym()
+{
+	int i;
+	char buf[30];
+
+	for (i=0; i<gensym; ++i) {
+		sprintf(buf, gs_name, i);
+		v_purge(buf);
+	}
+}
+
+
 /* hostname -- get system hostname or set my idea of the hostname */
 
 static int
@@ -242,7 +282,7 @@ run_hostname(argc, argv)
 	if (argc > 1) {
 		memtypes oval = stickymem;
 		stickymem = MEM_MALLOC;
-		myhostname = strsave(argv[1]);
+		myhostname = strdup(argv[1]);
 		stickymem = oval;
 	} else {
 		/* can't fail... */
@@ -315,7 +355,7 @@ run_erraddrlog(argc, argv)
 
 static conscell *
 run_dblookup(avl, il)
-	conscell *avl, *il;
+	conscell *avl, *il; /* Inputs gc protected */
 {
 	conscell *l;
 
@@ -331,151 +371,66 @@ run_dblookup(avl, il)
 
 static conscell *
 run_cadr(avl, il)
-	conscell *avl, *il;
+	conscell *avl, *il; /* Inputs gc protected */
 {
-	conscell *tmp;
-
 	il = cdar(avl);
 	if (il == NULL || STRING(il) || car(il) == NULL)
 		return NULL;
 	/* cdr */
-	/* setf preparation */
-	if (cdar(il)) {
-		il->prev = cdar(il)->prev;
-	} else if (car(il)->prev) {
-		if (car(il)->pflags)
-			il->prev = cdr(car(il)->prev);
-		else
-			il->prev = car(car(il)->prev);
-	}
-	il->pflags = 3;
 	car(il) = cdar(il);
 
 	/* car */
-	/* setf preparation */
-	if (car(il) == NULL) {
-		if (il->prev) {
-			il->prev = cdr(il->prev);
-			il->pflags = 0;
-		}
-		return il;
-	}
 	car(il) = copycell(car(il));	/* don't modify malloc'ed memory! */
 	cdar(il) = NULL;
-	car(il)->pflags |= 01 | 04;
 	return car(il);
 }
 
 static conscell *
 run_caddr(avl, il)
-	conscell *avl, *il;
+	conscell *avl, *il; /* Inputs gc protected */
 {
-	conscell *tmp;
-
 	il = cdar(avl);
 	if (il == NULL || STRING(il) || car(il) == NULL)
 		return NULL;
 	/* cdr */
-	/* setf preparation */
-	if (cdar(il)) {
-		il->prev = cdar(il)->prev;
-	} else if (car(il)->prev) {
-		if (car(il)->pflags)
-			il->prev = cdr(car(il)->prev);
-		else
-			il->prev = car(car(il)->prev);
-	}
-	il->pflags = 3;
 	car(il) = cdar(il);
 
 	/* cdr */
-	/* setf preparation */
-	if (cdar(il)) {
-		il->prev = cdar(il)->prev;
-	} else if (car(il)->prev) {
-		if (car(il)->pflags)
-			il->prev = cdr(car(il)->prev);
-		else
-			il->prev = car(car(il)->prev);
-	}
-	il->pflags = 3;
 	car(il) = cdar(il);
 
 	/* car */
 	/* setf preparation */
 	if (car(il) == NULL) {
-		if (il->prev) {
-			il->prev = cdr(il->prev);
-			il->pflags = 0;
-		}
 		return il;
 	}
 	car(il) = copycell(car(il));	/* don't modify malloc'ed memory! */
 	cdar(il) = NULL;
-	car(il)->pflags |= 01 | 04;
 	return car(il);
 }
 
 static conscell *
 run_cadddr(avl, il)
-	conscell *avl, *il;
+	conscell *avl, *il; /* Inputs gc protected */
 {
-	conscell *tmp;
-
 	il = cdar(avl);
 	if (il == NULL || STRING(il) || car(il) == NULL)
 		return NULL;
 	/* cdr */
-	/* setf preparation */
-	if (cdar(il)) {
-		il->prev = cdar(il)->prev;
-	} else if (car(il)->prev) {
-		if (car(il)->pflags)
-			il->prev = cdr(car(il)->prev);
-		else
-			il->prev = car(car(il)->prev);
-	}
-	il->pflags = 3;
 	car(il) = cdar(il);
 
 	/* cdr */
-	/* setf preparation */
-	if (cdar(il)) {
-		il->prev = cdar(il)->prev;
-	} else if (car(il)->prev) {
-		if (car(il)->pflags)
-			il->prev = cdr(car(il)->prev);
-		else
-			il->prev = car(car(il)->prev);
-	}
-	il->pflags = 3;
 	car(il) = cdar(il);
 
 	/* cdr */
-	/* setf preparation */
-	if (cdar(il)) {
-		il->prev = cdar(il)->prev;
-	} else if (car(il)->prev) {
-		if (car(il)->pflags)
-			il->prev = cdr(car(il)->prev);
-		else
-			il->prev = car(car(il)->prev);
-	}
-	il->pflags = 3;
 	car(il) = cdar(il);
 
 	/* car */
 	/* setf preparation */
 	if (car(il) == NULL) {
-		if (il->prev) {
-			il->prev = cdr(il->prev);
-			il->pflags = 0;
-		}
 		return il;
 	}
 	car(il) = copycell(car(il));	/* don't modify malloc'ed memory! */
 	cdar(il) = NULL;
-	car(il)->pflags |= 01 | 04;
 	return car(il);
 }
 
@@ -546,8 +501,8 @@ static int
 decmp(a, b)
      const void *a, *b;
 {
-	register const struct de *aa = (struct de *)a;
-	register const struct de *bb = (struct de *)b;
+	register const struct de *aa = (const struct de *)a;
+	register const struct de *bb = (const struct de *)b;
 
 	return bb->mtime - aa->mtime;
 }
@@ -576,6 +531,39 @@ rd_endstability()
 		free((char *)nbarray);
 }
 
+static int
+run_doit(argc, argv)
+	int argc;
+	const char *argv[];
+{
+	const char *filename;
+	char *sh_memlevel = getlevel(MEM_SHCMD);
+	int r;
+	const char *av[3];
+
+	if (argc != 2) {
+	  fprintf(stderr, "Usage: %s <filename>\n", argv[0]);
+	  return 1;
+	}
+
+	filename = argv[1];
+
+	/* Do one file, return value is 0 or 1,
+	   depending on actually doing something
+	   on a file */
+
+	gensym = 1;
+	av[0] = "process"; /* I think this needs to be here */
+	av[1] = filename;
+	av[2] = NULL;
+	r = s_apply(2, av); /* "process" filename (within  rd_doit() ) */
+	free_gensym();
+
+	setlevel(MEM_SHCMD,sh_memlevel);
+
+	return r;
+}
+
 static int rd_doit __((const char *filename, const char *dirs));
 static int
 rd_doit(filename, dirs)
@@ -613,9 +601,12 @@ rd_doit(filename, dirs)
 	if (p != NULL) {
 	  /* message file is "inode-pid" */
 	  thatpid = atoi(p+1);
+
+#if 0 /* very old thing, may harm at Solaris 2.6 ! */
 	  if (thatpid < 10) {	/* old-style locking.. */
 	    thatpid = 0;
 	  }
+#endif
 	  /* Probe it!
 	     Does the process exist ? */
 	  if (thatpid && kill(thatpid,0)==0 && thatpid != router_id) {
@@ -637,12 +628,14 @@ rd_doit(filename, dirs)
 	     */
 	  }
 	}
-	if (p == NULL || thatpid != router_id) {
-	  /* Not already in format of 'inode-pid' */
+	if (strncmp(filename,"core",4) != 0 &&
+	    (p == NULL || thatpid != router_id)) {
+	  /* Not a core file, and ...
+	     not already in format of 'inode-pid' */
 	  /* If the pid did exist, we do not touch on that file,
 	     on the other hand, we need to rename the file now.. */
 #ifdef	USE_ALLOCA
-	  buf = alloca(len+16);
+	  buf = (char*)alloca(len+16);
 #else
 	  if (blen == 0) {
 	    blen = len+16;
@@ -709,7 +702,7 @@ rd_stability(dirp,dirs)
 		/* Handle only files beginning with number -- plus "core"-
 		   files.. */
 		if (!(dp->d_name[0] >= '0' && dp->d_name[0] <= '9') &&
-		    strcmp(dp->d_name,"core") != 0)
+		    strncmp(dp->d_name,"core",4) != 0)
 			continue;
 
 		/* See that the file is a regular file! */
@@ -760,6 +753,14 @@ rd_stability(dirp,dirs)
 		if (gothup) 
 			dohup(0);
 		did_cnt += rd_doit(nbarray + dearray[deindex].f_name, dirs);
+
+		/* Maybe only process few files out of the low-priority
+		   subdirs, so we can go back and see if any higher-priority
+		   jobs have been created */
+		if ((*dirs) &&
+		    ((routerdirloops) && (routerdirloops == did_cnt)))
+		  break;
+
 	}
 	return did_cnt;
 }
@@ -785,7 +786,7 @@ rd_instability(dirp, dirs)
 		/* Handle only files beginning with number -- plus "core"-
 		   files.. */
 		if (!(dp->d_name[0] >= '0' && dp->d_name[0] <= '9') &&
-		    strcmp(dp->d_name,"core") != 0)
+		    strncmp(dp->d_name,"core",4) != 0)
 			continue;
 
 		/* See that the file is a regular file! */
@@ -794,6 +795,13 @@ rd_instability(dirp, dirs)
 		if (!S_ISREG(statbuf.st_mode)) continue; /* Hmm..  */
 
 		did_cnt += rd_doit(dp->d_name, dirs);
+
+		/* Only process one file out of the low-priority subdirs,
+		   so we can go back and see if any higher-priority
+		   jobs have been created */
+		if (*dirs)
+			break;
+
 	}
 	return did_cnt;
 }
@@ -837,6 +845,8 @@ run_daemon(argc, argv)
 	memtypes oval = stickymem;
 	struct stat stb;
 
+	router_id = getpid();
+
 	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* mustexit = 1 */
 	for (i=0; i<ROUTERDIR_CNT; ++i) {
 		dirp[i] = NULL; dirs[i] = NULL;
@@ -864,7 +874,7 @@ run_daemon(argc, argv)
 			  dirp[i]->dd_size = 0;
 #endif
 #endif
-			  dirs[i] = strsave(pathbuf);
+			  dirs[i] = strdup(pathbuf);
 			  ++i;
 			}
 			if (s)
@@ -881,7 +891,7 @@ run_daemon(argc, argv)
 	stickymem = oval;
 	did_cnt = 0;
 	i = -1;
-	for (;;) {
+	for (; !mustexit ;) {
 		++i;	/* Increment it */
 		/* The last of the alternate dirs ?  Reset.. */
 		if (i >= ROUTERDIR_CNT || dirs[i] == NULL) {
@@ -948,6 +958,7 @@ run_daemon(argc, argv)
 		close(dirp[i]->dd_fd);
 #endif
 		closedir(dirp[i]);
+		dirp[i] = NULL;
 #endif
 
 		if (mustexit)
@@ -1005,9 +1016,9 @@ run_process(argc, argv)
 		return PERR_USAGE;
 	}
 #ifdef	USE_ALLOCA
-	file = alloca(strlen(argv[1])+1);
+	file = (char*)alloca(strlen(argv[1])+1);
 #else
-	file = emalloc(strlen(argv[1])+1);
+	file = (char*)emalloc(strlen(argv[1])+1);
 #endif
 	strcpy(file, argv[1]);
 
@@ -1279,8 +1290,10 @@ run_listexpand(avl, il)
 	struct envelope *e;
 	struct address *ap, *aroot = NULL, **atail = &aroot;
 	token822 *t;
-	conscell *al, *alp = NULL, *tmp;
-	char *localpart, *origaddr, *attributenam;
+	conscell *al = NULL, *alp = NULL, *tmp = NULL;
+	conscell *plustail = NULL, *domain = NULL;
+	conscell *l, *lrc;
+	char *localpart = NULL, *origaddr = NULL, *attributenam = NULL;
 	int c, n, errflag, stuff;
 	volatile int cnt;
 	const char *comment, *erroraddress;
@@ -1297,6 +1310,7 @@ run_listexpand(avl, il)
 	int okaddresses = 0;
 	int errcount = 0;
 	int linecnt = 0;
+	GCVARS3;
 
 	il = cdar(avl); /* The CDR (next) of the arg list.. */
 
@@ -1352,10 +1366,10 @@ run_listexpand(avl, il)
 	tmp = il;
 	for (; tmp != NULL; tmp = cdr(tmp)) ++cnt; /* Count arguments.. */
 
-	if (errflag || cnt != 3 ||
+	if (errflag || cnt < 3 || cnt > 5 ||
 	    !STRING(il) || !STRING(cdr(il)) || !STRING(cddr(il)) ) {
 		fprintf(stderr,
-			"Usage: %s [ -e error-address ] [ -E errors-to-address ] [-p privilege] [ -c comment ] [ -N notarystring ] $attribute $localpart $origaddr < /file/path \n",
+			"Usage: %s [ -e error-address ] [ -E errors-to-address ] [-p privilege] [ -c comment ] [ -N notarystring ] $attribute $localpart $origaddr [$plustail [$domain]]< /file/path \n",
 		car(avl)->string);
 		if (errors_to != olderrors)
 		  free(errors_to);
@@ -1366,6 +1380,12 @@ run_listexpand(avl, il)
 	attributenam = (char*)    (il)->string;
 	localpart    = (char*) cdr(il)->string;
 	origaddr     = (char*) cddr(il)->string;
+	if (cdr(cddr(il))) {
+	  plustail = cdr(cddr(il));
+	  if (cddr(cddr(il)))
+	    domain = cddr(cddr(il));
+	}
+	
 
 	/* We (memory-)leak this stuff for a moment.. (but it is tmalloc()ed)*/
 	e = (struct envelope *)tmalloc(sizeof (struct envelope));
@@ -1377,7 +1397,7 @@ run_listexpand(avl, il)
 	hs.h_descriptor = &aliashdr;
 	hs.h_pname = (char*) car(avl)->string;
 
-	initline(4096);
+	initzline(4096);
 
 	/*
 	 * These hoops are so we can do this for both real stdin and
@@ -1400,18 +1420,18 @@ run_listexpand(avl, il)
 	/* use a constant buffer to avoid memory leaks from stdio usage */
 	setvbuf(fp, buf, _IOFBF, sizeof buf);
 	c = 0;
-	while ((n = getline(fp)) > 0) {
+	while ((n = zgetline(fp)) > 0) {
 		++linecnt;
 		/*
 		 * For sendmail compatibility, addresses may not cross line
 		 * boundaries, and line boundary is just as good an address
 		 * separator as comma is, sigh.
 		 */
-		if (linebuf[n-1] == '\n')
+		if (zlinebuf[n-1] == '\n')
 			--n;
 		stuff = 0;
-		s_end = &linebuf[n];
-		for (s = linebuf; s < s_end; ++s) {
+		s_end = & zlinebuf[n];
+		for (s = zlinebuf; s < s_end; ++s) {
 			if (isascii(*s) && !isspace(*s))
 				c = stuff = 1;
 			if (*s == ',')
@@ -1420,7 +1440,7 @@ run_listexpand(avl, il)
 		if (!stuff)
 			continue;
 		if (c) {
-			linebuf[n] = ',';
+			zlinebuf[n] = ',';
 			++n;
 		}
 
@@ -1431,8 +1451,8 @@ run_listexpand(avl, il)
 		 * This is just to not to force people to edit their
 		 * .forwards :-(  "\haa" works as expected  :-)
 		 **/
-		if (linebuf[0] == '\\') linebuf[0] = ' ';
-		for (s = linebuf+1; s < s_end; ++s) {
+		if (zlinebuf[0] == '\\') zlinebuf[0] = ' ';
+		for (s = zlinebuf+1; s < s_end; ++s) {
 		  if (s[-1] == ',' && s[0]=='\\') s[0] = ' ';
 		}
 		/* end of sendmail compatibility kluge */
@@ -1444,7 +1464,7 @@ run_listexpand(avl, il)
 		 * so errors will print out nicely.
 		 */
 
-		hs.h_lines = t = makeToken(linebuf, n);
+		hs.h_lines = t = makeToken(zlinebuf, n);
 		t->t_type = Line;
 
 		/* fix up any trailing comma (more sendmail
@@ -1549,11 +1569,13 @@ run_listexpand(avl, il)
 	}
 
 	cnt = 0;
-	al = NULL;
+
+	al = l = lrc = NULL;
+	GCPRO3(al, l, lrc);
+
 	for (ap = aroot; ap != NULL; ap = ap->a_next) {
-		int rc;
-		int omem;
-		conscell *l, *lrc;
+		int rc, slen;
+		memtypes omem;
 		char *s2, *se;
 
 		buf[0] = 0;
@@ -1599,9 +1621,27 @@ run_listexpand(avl, il)
 		omem = stickymem;
 		/* stickymem = MEM_MALLOC; */
 
-		l       = newstring(strsave(buf));
-		cdr(l)  = newstring(strsave(origaddr));
-		cddr(l) = newstring(strsave(s));
+		/* The set of parameters for the rrouter() script
+		   function are:
+		   - address
+		   - origaddress
+		   - Attribute variable name
+		   - plustail
+		   - domain
+		   (The last two were added in June-1998)
+		*/
+
+		slen = strlen(buf);
+		l         = newstring(dupnstr(buf, slen), slen);
+		slen = strlen(origaddr);
+		cdr(l)    = newstring(dupnstr(origaddr, slen), slen);
+		slen = strlen(s);
+		cddr(l)   = newstring(dupnstr(s, slen), slen);
+		if (plustail != NULL) {
+		  cdr(cddr(l))  = s_copy_chain(plustail);
+		  if (domain != NULL)
+		    cddr(cddr(l)) = s_copy_chain(domain);
+		}
 		l = ncons(l);
 
 		stickymem = omem;
@@ -1623,11 +1663,12 @@ run_listexpand(avl, il)
 #endif
 #if 0 /* XX: This HOLD handling didn't work ?? */
 		if (deferit && (d = v_find(DEFER))) {
-		  s_free_tree(lrc);
+		  /* s_free_tree(lrc); */
 		  lrc = NULL;
-		  l = conststring("hold");
+		  l = conststring("hold", 4);
 		  cdr(l)   = copycell(cdr(d));
-		  cddr(l)  = newstring(strsave(buf));
+		  slen = strlen(buf);
+		  cddr(l)  = newstring(dupnstr(buf, slen), slen);
 		  cdddr(l) = car(attributes);
 		  l = ncons(l);
 		  l = ncons(l);
@@ -1637,7 +1678,7 @@ run_listexpand(avl, il)
 		if (rc != 0 || lrc == NULL || !LIST(lrc)) {
 		  /* $(rrouter xx xx xx)  returned something invalid.. */
 #ifdef use_lapply
-		  s_free_tree(lrc);
+		  /* s_free_tree(lrc); */
 #endif
 		  lrc = NULL;
 		  continue;
@@ -1658,23 +1699,24 @@ run_listexpand(avl, il)
 			      progname, ROUTER);
 		      s_grind(lrc, stderr);
 #ifdef use_lapply
-		      s_free_tree(lrc);
+		      /* s_free_tree(lrc); */
 #endif
 		      lrc = NULL;
 		      if (errors_to != olderrors)
 			free(errors_to);
 		      errors_to = olderrors;
+		      UNGCPRO3;
 		      return NULL;
 		    }
-		    l = s_copy_tree(lrc);
+		    l = s_copy_chain(lrc);
 		  } else {
-		    l = s_copy_tree(lrc);
+		    l = s_copy_chain(lrc);
 		    l = ncons(l);
 		    l = ncons(l);
 		  }
 
 #ifdef use_lapply
-		  s_free_tree(lrc);
+		  /* s_free_tree(lrc); */
 #endif
 		  lrc = NULL;
 		}
@@ -1710,12 +1752,15 @@ run_listexpand(avl, il)
 	}
 
 	if (al == NULL) { /* ERROR! NO ADDRESSES! */
-	  al = conststring("error");
-	  cdr(al)  = conststring("expansion");
-	  cddr(al) = newstring(strsave(localpart));
+	  int slen;
+	  al = conststring("error", 5);
+	  cdr(al)  = conststring("expansion", 9);
+	  slen = strlen(localpart);
+	  cddr(al) = newstring(dupnstr(localpart, slen), slen);
 	  al = ncons(al);
 	  al = ncons(al);
 	}
+	UNGCPRO3;
 
 	if (errors_to != olderrors)
 	  free(errors_to);
@@ -1790,7 +1835,7 @@ run_listaddrs(argc, argv)
 	hs.h_descriptor = &aliashdr;
 	hs.h_pname = argv[0];
 
-	initline(4096);
+	initzline(4096);
 
 	/*
 	 * These hoops are so we can do this for both real stdin and
@@ -1813,18 +1858,18 @@ run_listaddrs(argc, argv)
 	/* use a constant buffer to avoid memory leaks from stdio usage */
 	setvbuf(fp, buf, _IOFBF, sizeof buf);
 	c = 0;
-	while ((n = getline(fp)) > 0) {
+	while ((n = zgetline(fp)) > 0) {
 		++linecnt;
 		/*
 		 * For sendmail compatibility, addresses may not cross line
 		 * boundaries, and line boundary is just as good an address
 		 * separator as comma is, sigh.
 		 */
-		if (linebuf[n-1] == '\n')
+		if (zlinebuf[n-1] == '\n')
 			--n;
 		stuff = 0;
-		s_end = &linebuf[n];
-		for (s = linebuf; s < s_end; ++s) {
+		s_end = &zlinebuf[n];
+		for (s = zlinebuf; s < s_end; ++s) {
 			if (isascii(*s) && !isspace(*s))
 				c = stuff = 1;
 			if (*s == ',')
@@ -1833,7 +1878,7 @@ run_listaddrs(argc, argv)
 		if (!stuff)
 			continue;
 		if (c) {
-			linebuf[n] = ',';
+			zlinebuf[n] = ',';
 			++n;
 		}
 
@@ -1844,8 +1889,8 @@ run_listaddrs(argc, argv)
 		 * This is just to not to force people to edit their
 		 * .forwards :-(  "\haa" works as expected  :-)
 		 **/
-		if (linebuf[0] == '\\') linebuf[0] = ' ';
-		for (s = linebuf+1; s < s_end; ++s) {
+		if (zlinebuf[0] == '\\') zlinebuf[0] = ' ';
+		for (s = zlinebuf+1; s < s_end; ++s) {
 		  if (s[-1] == ',' && s[0]=='\\') s[0] = ' ';
 		}
 		/* end of sendmail compatibility kluge */
@@ -1856,7 +1901,7 @@ run_listaddrs(argc, argv)
 		 * so errors will print out nicely.
 		 */
 
-		hs.h_lines = t = makeToken(linebuf, n);
+		hs.h_lines = t = makeToken(zlinebuf, n);
 		t->t_type = Line;
 
 		/* fix up any trailing comma (more sendmail
@@ -2079,9 +2124,9 @@ run_filepriv(argc, argv)
 	} else if (cp == file)	/* root path */
 		++cp;
 #ifdef	USE_ALLOCA
-	dir = alloca(cp - file + 1);
+	dir = (char*)alloca(cp - file + 1);
 #else
-	dir = emalloc(cp - file + 1);
+	dir = (char*)emalloc(cp - file + 1);
 #endif
 	memcpy(dir, file, cp - file);
 	dir[cp - file] = '\0';
@@ -2148,30 +2193,6 @@ run_runas(argc, argv)
 	return r;
 }
 
-int gensym;
-const char *gs_name = "g%d";
-
-static int
-run_gensym(argc, argv)
-	int argc;
-	const char *argv[];
-{
-	printf(gs_name, gensym++);
-	putchar('\n');
-	return 0;
-}
-
-static void
-free_gensym()
-{
-	int i;
-	char buf[30];
-
-	for (i=0; i<gensym; ++i) {
-		sprintf(buf, gs_name, i);
-		v_purge(buf);
-	}
-}
 
 static int
 run_uid2login(argc, argv)
@@ -2229,6 +2250,53 @@ run_basename(argc, argv)
 	return 0;
 }
 
+static int
+run_syslog(argc, argv)
+	int argc;
+	const char *argv[];
+{
+	int c;
+	int prio = LOG_INFO;
+	int errflg = 0;
+	optind = 1;
+
+	while ((c = getopt(argc, (char*const*)argv, "p:")) != EOF) {
+		switch (c) {
+		case 'p':	/* priority */
+			if(!strcmp(optarg, "debug")) {
+				prio = LOG_DEBUG;
+			} else if(!strcmp(optarg, "info")) {
+				prio = LOG_INFO;
+			} else if(!strcmp(optarg, "notice")) {
+				prio = LOG_NOTICE;
+			} else if(!strcmp(optarg, "warning")) {
+				prio = LOG_WARNING;
+			} else if(!strcmp(optarg, "err")) {
+				prio = LOG_ERR;
+			} else if(!strcmp(optarg, "crit")) {
+				prio = LOG_CRIT;
+			} else if(!strcmp(optarg, "alert")) {
+				prio = LOG_ALERT;
+			} else if(!strcmp(optarg, "emerg")) {
+				prio = LOG_EMERG;
+			} else {
+				++errflg;
+			}
+			break;
+		default:
+			++errflg;
+			break;
+		}
+	}
+
+	if (errflg || optind != argc - 1) {
+		fprintf(stderr, "Usage: %s [-p prio] string\n", argv[0]);
+		return 1;
+	}
+	zsyslog((prio, "%s", argv[optind]));
+	return 0;
+}
+
 static int
 run_recase(argc, argv)
 	int argc;
@@ -2417,8 +2485,8 @@ run_dequote(argc, argv)
 	len = strlen(s);
 
 	if (len > 1 &&
-	    (*s == '"'  && s[len-1] == '"' ) ||
-	    (*s == '\'' && s[len-1] == '\'')) {
+	    ((*s == '"'  && s[len-1] == '"' ) ||
+	     (*s == '\'' && s[len-1] == '\''))) {
 	  fwrite(s+1,1,len-2,stdout);
 	} else
 	  fwrite(s,1,len,stdout);
@@ -2434,7 +2502,6 @@ run_condquote(argc, argv)
 {
 	int len;
 	const char *s = argv[1];
-	const char *sp, *bs;
 	int mustquote = 0;
 	int candequote = 0;
 	int c;
@@ -2450,8 +2517,9 @@ run_condquote(argc, argv)
 	  ++s; /* Starting quote */
 	  while (*s && *s != '"') {
 	    /* While within quoted string */
-	    if (*s == '\\')
+	    if (*s == '\\') {
 	      ++s;
+	    }
 	    if (*s != 0)
 	      ++s;
 	  }
@@ -2464,6 +2532,8 @@ run_condquote(argc, argv)
 	  if (c == '\\') {
 	    ++s;
 	    c = *s;
+	    if (c == '!') /* fooo\!baar */
+	      mustquote = 1;
 	  } else if (c == ' ' || c == '\t')
 	    mustquote = 1; /* Unquoted spaces! */
 	  if (c != 0)
@@ -2474,7 +2544,7 @@ run_condquote(argc, argv)
 	len = strlen(s);
 
 	/* Quoted, and without a need for quotes */
-	if (candequote) {
+	if (!mustquote && candequote) {
 	  /* XXX: THIS SHOULD REALLY USE SOME SYNTAX SCANNER -- LIKE THAT ONE
 	          FOR SMTP: RFC821SCN()    */
 	  fwrite(argv[1] + 1, 1, len -2, stdout); /* Dequoted! */
