@@ -36,7 +36,7 @@ open_core(sip)
 		return NULL;
 	if (spt->symbols == NULL)
 		spt->symbols = sp_init();
-	if (spt == NULL)  /* Failed, damn! */
+	if (spt->symbols == NULL)  /* Failed, damn! */
 		return NULL;
 	return spt;
 }
@@ -67,8 +67,9 @@ search_core(sip)
 {
 	struct sptree *db;
 	struct spblk *spl;
-	conscell *tmp;
 	spkey_t spk;
+	char *s;
+	int slen;
 
 	db = open_core(sip);
 	if (db == NULL)
@@ -80,8 +81,10 @@ search_core(sip)
 	if (spl == NULL)
 		return NULL;
 	if (spl->data == NULL)
-		return newstring(strnsave("", 1));
-	return newstring(strsave((char *)spl->data));
+		return conststring("", 0);
+	slen = strlen((const char *)spl->data);
+	s = dupnstr((const char *)spl->data, slen);
+	return newstring(s, slen);
 }
 
 /*
@@ -99,8 +102,9 @@ icfreedata(spl)
 }
 
 void
-close_core(sip)
+close_core(sip,comment)
 	search_info *sip;
+	const char *comment;
 {
 	struct sptree *db;
 
@@ -161,19 +165,15 @@ add_core(sip, value)
 	struct sptree *db;
 	struct spblk *spl;
 	spkey_t spk;
-	memtypes oval;
 
 	db = open_core(sip);
 	if (db == NULL)
 		return EOF;
 
-	oval = stickymem;
-	stickymem = MEM_MALLOC;
 	if (value == NULL || *value == '\0')
 		value = NULL;
 	else
-		value = strsave(value);
-	stickymem = oval;
+		value = strdup(value);
 
 	spk = symbol_db(sip->key, db->symbols);
 	spl = sp_lookup(spk, db);
@@ -229,9 +229,10 @@ icprintNS(spl)
 	struct spblk *spl;
 {
 	if (spl->data != NULL)
-		fprintf(pcfp, "%d\t%s\n", (int)spl->key, (char *)spl->data);
+		fprintf(pcfp, "%d\t%s\n",
+			(int)spl->key, (const char *)spl->data);
 	else
-		fprintf(pcfp, "%d\n", spl->key);
+		fprintf(pcfp, "%d\n", (int) spl->key);
 	return 0;
 }
 
@@ -250,9 +251,11 @@ icprintSS(spl)
 	struct spblk *spl;
 {
 	if (spl->data != NULL)
-		fprintf(pcfp, "%s\t%s\n", pname(spl->key), (char *)spl->data);
+		fprintf(pcfp, "%s\t%s\n",
+			pname(spl->key), (const char *)spl->data);
 	else
-		fprintf(pcfp, "%s\n", pname(spl->key));
+		fprintf(pcfp, "%s\n",
+			pname(spl->key));
 	return 0;
 }
 
