@@ -15,14 +15,14 @@
 STATIC		/* This will apply to the array in the .sst file */
 #include "rfc822.sst.c"
 
-extern void	EmitToken __((struct token *t, ComponentClass ac, struct address *ap));
+extern void	EmitToken __((token822 *t, ComponentClass ac, struct address *ap));
 STATIC const char * Input822TokenName __((int));
 
 /* private semantics mechanism variables: */
 STATIC ComponentClass currentTokenType;
-STATIC struct token *ptrNextInputToken, *ptrAcceptedToken;
-STATIC struct token *headPending, **pending;
-STATIC struct token *headDeferred, **deferred;
+STATIC token822 *ptrNextInputToken, *ptrAcceptedToken;
+STATIC token822 *headPending, **pending;
+STATIC token822 *headDeferred, **deferred;
 
 /* Table Walker State */
 STATIC int	processing;		/* are we running the table walker? */
@@ -33,7 +33,7 @@ STATIC TableOperation	operation;	/* the current operation */
 /* convenience */
 STATIC struct address nullAddr = { NULL, NULL, newAddress, 0, NULL, NULL };
 #ifdef ATHACK
-STATIC struct token atsigntoken = { "@", 1, Special, NULL };
+STATIC token822 atsigntoken = { "@", 1, Special, NULL };
 #endif
 
 /* Tracing Control */
@@ -151,7 +151,7 @@ SslError(errCode, ap)
 	ErrorCodes errCode;
 	struct address *ap;
 {
-	struct token *t;
+	token822 *t;
 	int i;
 
 	if (errCode == eNoError)
@@ -183,10 +183,10 @@ SslError(errCode, ap)
  * indicators and the values and text associated with input tokens.  
  */
 
-STATIC void AcceptInputToken __((struct token **, struct address *));
+STATIC void AcceptInputToken __((token822 **, struct address *));
 STATIC void
 AcceptInputToken(tlistp, ap)
-	struct token **tlistp;
+	token822 **tlistp;
 	struct address *ap;
 {
 
@@ -345,10 +345,10 @@ ok:
  * been matched.
  */
 
-STATIC void SslSyntaxError __((struct token **, struct address *));
+STATIC void SslSyntaxError __((token822 **, struct address *));
 STATIC void
 SslSyntaxError(tlistp, ap)
-	struct token **tlistp;
+	token822 **tlistp;
 	struct address *ap;
 {
 	if (operation != oInput && operation != oInputAny)
@@ -538,11 +538,11 @@ SslChoice(choiceTag)
 
 void
 EmitToken(t, ac, ap)
-	struct token *t;
+	token822 *t;
 	ComponentClass ac;
 	struct address *ap;
 {
-	struct token *nt, **ptp;
+	token822 *nt, **ptp;
 	struct addr *na;
 	AddrComponent type;
 
@@ -618,7 +618,7 @@ struct address *
 revaddress(ap)
 	struct address *ap;
 {
-	register struct token	*rprev, *rnext, *r;
+	register token822	*rprev, *rnext, *r;
 	register struct addr	*pprev, *pnext, *p;
 
 	/* reverse order of address components and tokens */
@@ -642,14 +642,14 @@ revaddress(ap)
 	return ap;
 }
 
-STATIC void revappend __((struct token **, struct address *));
+STATIC void revappend __((token822 **, struct address *));
 STATIC void
 revappend(tprev,ap)
-struct token **tprev;
+token822 **tprev;
 struct address *ap;
 {
 	struct addr *p;
-	struct token *t;
+	token822 *t;
 
 	ap = revaddress(ap);
 	/* append all the tokens together */
@@ -685,12 +685,12 @@ struct address *ap;
 union misc
 parse822(entry, tlistp, ltmp, tfp)
 	HeaderSemantics	entry;
-	struct token **tlistp;
+	token822 **tlistp;
 	struct tm *ltmp;
 	FILE *tfp;
 {
 	int		inReceived, i;
-	struct token	*t, *torig;
+	token822	*t, *torig;
 	struct address	*ap, *na, *pap;
 	struct received	rcvd;
 	union misc	retval;
