@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1995
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
  */
 
 /*
@@ -22,6 +22,7 @@
 #include "mail.h"
 #include <string.h>
 #include "ta.h"
+#include "sysexits.h"
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -50,6 +51,7 @@
 #include "prototypes.h"
 #include "zsyslog.h"
 #include "libz.h"
+#include <grp.h>
 
 extern int optind;
 extern char *optarg;
@@ -78,7 +80,7 @@ struct conshell *envarlist = NULL;
 				  need open -- or any of its children.. */
 int	transportmaxnofiles = TRANSPORTMAXNOFILES; /* Default value */
 const char * progname;
-const char * postoffice;
+extern const char * postoffice;
 const char * rendezvous;
 const char * pidfile = PID_SCHEDULER;
 const char * mailshare;
@@ -118,6 +120,7 @@ char * procselect = NULL;	/* Non-null defines  channel/host specifier
 				   job-specifier deletions. */
 char * procselhost = NULL;	/* Just spliced out 'host'-part of the above */
 extern int forkrate_limit;	/* How many forks per second ? */
+int	mailqmode = 1;		/* ZMailer v1.0 mode on mailq */
 
 static int vtxprep_skip      = 0;
 static int vtxprep_skip_any  = 0;
@@ -126,15 +129,16 @@ static time_t next_dirscan     = 0;
 static time_t next_idlecleanup = 0;
 static struct sptree *dirscan_mesh = NULL;
 static int newents_limit = 200;
+extern int never_full_content; /* on conf.c */
 
 #include "memtypes.h"
 extern memtypes stickymem;
 
-static struct ctlfile *schedule __((int fd, const char *file, struct config_entry *ce, long ino));
-static struct ctlfile *vtxprep __((struct ctlfile *, const char *));
+static struct ctlfile *schedule __((int fd, const char *file, long ino, const int));
+static struct ctlfile *vtxprep __((struct ctlfile *, const char *, const int));
 static int  vtxmatch __((struct vertex *, struct config_entry *));
 static void link_in __((int flag, struct vertex *vp, const char *s));
-static int  lockverify __((struct ctlfile *, const char *));
+static int  lockverify __((struct ctlfile *, const char *, const int));
 static int  globmatch   __((const char *, const char*));
 static void vtxdo   __((struct vertex *, struct config_entry *, const char *));
 
@@ -149,6 +153,10 @@ static RETSIGTYPE sig_readcf __((int sig));
 
 extern char *strerror __((int err));
 
+static struct MIB_MtaEntry MIBMtaEntryLocal = {0,};
+struct MIB_MtaEntry *MIBMtaEntry = &MIBMtaEntryLocal;
+
+
 static int    timeserver_pid = 0;
 extern time_t mytime          __((time_t *));
 static void   init_timeserver __((void));
@@ -193,10 +201,10 @@ int sig;
 	    fprintf(stderr,"%s: cannot open statuslog: %s, errno=%d\n", progname, statusfn, errno);
 	    return -1;
 	  }
+	  setvbuf(statuslog, (char *)NULL, _IOLBF, 0);
 #if defined(F_SETFD)
 	  fcntl(FILENO(statuslog), F_SETFD, 1); /* close-on-exec */
 #endif
-	  setvbuf(statuslog, (char *)NULL, _IOLBF, 0);
 	}
 	SIGNAL_HANDLE(SIGHUP, (RETSIGTYPE(*)__((int))) loginitsched);
 	return 0;
@@ -219,14 +227,14 @@ struct dirqueue {
 	struct dirstatname **stats2;
 };
 
-static int dirqueuescan __((const char *dir,struct dirqueue *dq,int subdirs));
-static int syncweb __((struct dirqueue *dq, struct config_entry *ce));
+static int dirqueuescan __((const char *dir,struct dirqueue *dq, int subdirs));
+static int syncweb __((struct dirqueue *dq));
 
 int global_maxkids = 1000;
 time_t now;
 
 FILE * vfp_open __((struct ctlfile *));
-FILE *vfp_open(cfp)
+FILE * vfp_open(cfp)
 struct ctlfile *cfp;
 {
 	FILE *vfp;
@@ -236,27 +244,22 @@ struct ctlfile *cfp;
 	setreuid(0, 0);
 	if (!vfp) return NULL; /* Failure to open */
 
-	fseek(vfp, (off_t)0, SEEK_END);
 	setvbuf(vfp, (char *)NULL, _IOLBF, 0);   /* Set LINEBUFFERING */
+	fseek(vfp, (off_t)0, SEEK_END);
 	return vfp;
 }
 
 
 static void cfp_free __((struct ctlfile *cfp, struct spblk *spl));
-static void cfp_free(cfp, spl)
+static void cfp_free0 __((struct ctlfile *cfp));
+
+static void cfp_free0(cfp)
 struct ctlfile *cfp;
-struct spblk   *spl;
 {
+
 	struct vertex *vp, *nvp;
 
-	/* Delete from the  spt_mesh[]  */
-
-	if (spl == NULL)
-	  spl = sp_lookup((u_long)(cfp->id), spt_mesh[L_CTLFILE]);
-	if (spl != NULL)
-	  sp_delete(spl, spt_mesh[L_CTLFILE]);
-
-	/* And from memory */
+	/* Delete from memory */
 
 	if (cfp->vfpfn) {
 	  FILE *vfp = vfp_open(cfp);
@@ -271,6 +274,7 @@ struct spblk   *spl;
 	if (cfp->head != NULL) {
 	  for (vp = cfp->head; vp != NULL; vp = nvp) {
 	    nvp = vp->next[L_CTLFILE];
+	    MIBMtaEntry->mtaStoredRecipients -= vp->ngroup;
 	    vp->ngroup = 0;
 	    unvertex(vp,-1,1); /* Don't unlink()! Just free()! */
 	  }
@@ -279,6 +283,23 @@ struct spblk   *spl;
 	}
 }
 
+static void cfp_free(cfp, spl)
+struct ctlfile *cfp;
+struct spblk   *spl;
+{
+	/* Delete from the  spt_mesh[]  */
+
+	if (spl == NULL)
+	  spl = sp_lookup((u_long)(cfp->id), spt_mesh[L_CTLFILE]);
+	if (spl != NULL)
+	  sp_delete(spl, spt_mesh[L_CTLFILE]);
+
+	/* And from the memory */
+
+	cfp_free0(cfp);
+}
+
+
 static int ctl_free __((struct spblk *spl));
 static int ctl_free(spl)
 struct spblk *spl;
@@ -313,6 +334,9 @@ const char **ArgvSave;
 
 extern int main __((int, const char **));
 
+static struct config_entry *cehead = NULL;
+
+
 int
 main(argc, argv)
 	int argc;
@@ -322,7 +346,6 @@ main(argc, argv)
 	char *config, *cp;
 	int i, daemonflg, c, errflg, version, fd;
 	long offout, offerr;
-	struct config_entry *tp;
 
 	char *t, *syslogflg = getzenv("SYSLOGFLG");
 	if (syslogflg == NULL)
@@ -370,7 +393,7 @@ main(argc, argv)
 	verbose = errflg = version = 0;
 	for (;;) {
 		c = getopt(argc, (char*const*)argv,
-			   "divE:f:Fl:HL:N:p:P:q:QR:SVW");
+			   "divE:f:Fl:HL:M:nN:p:P:q:QR:SVW");
 		if (c == EOF)
 		  break;
 		switch (c) {
@@ -392,11 +415,11 @@ main(argc, argv)
 			  perror("Can't open statistics log file (-l)");
 			  exit(1);
 			}
+			setvbuf(statuslog, (char *)NULL, _IOLBF, 0);
 #if defined(F_SETFD)
 			fcntl(FILENO(statuslog), F_SETFD, 1);
 			/* close-on-exec */
 #endif
-			setvbuf(statuslog, (char *)NULL, _IOLBF, 0);
 			break;
 		case 'H':
 			if (hashlevels < 2)
@@ -405,6 +428,16 @@ main(argc, argv)
 		case 'L':	/* override default log file */
 			logfn = optarg;
 			break;
+		case 'M':
+			mailqmode = atoi(optarg);
+			if (mailqmode < 1 || mailqmode > 2) {
+			  fprintf(stderr,"scheduler: -M parameter is either 1, or 2\n");
+			  exit(EX_USAGE);
+			}
+			break;
+		case 'n':
+			never_full_content = !never_full_content;
+			break;
 		case 'N':
 			if ((transportmaxnofiles = atoi(optarg)) < 10)
 				transportmaxnofiles = TRANSPORTMAXNOFILES;
@@ -463,7 +496,7 @@ main(argc, argv)
 
 	if (errflg) {
 	  fprintf(stderr,
-		  "Usage: %s [-dHisvV -f configfile -L logfile -P postoffice -Q rendezvous]\n",
+		  "Usage: %s [-dHisvV -M (1|2) -f configfile -L logfile -P postoffice -Q rendezvous]\n",
 		  progname);
 	  exit(128+errflg);
 	}
@@ -519,8 +552,8 @@ main(argc, argv)
 				       + strlen(qcf_suffix)));
 	  sprintf(config, "%s/%s.%s", mailshare, progname, qcf_suffix);
 	}
-	tp = readconfig(config);
-	if (tp == NULL) {
+	cehead = readconfig(config);
+	if (cehead == NULL) {
 	  cp = emalloc(strlen(config)+50);
 	  sprintf(cp, "null control file, propably errors in it: %s", config);
 	  die(1, cp);
@@ -549,7 +582,7 @@ main(argc, argv)
 	  detach();		/* leave worldy matters behind */
 	  mytime(&now);
 	  printf("%s: scheduler daemon (%s)\n\tpid %d started at %s\n",
-		 progname, Version, getpid(), (char *)rfc822date(&now));
+		 progname, Version, (int)getpid(), (char *)rfc822date(&now));
 	}
 	/* Actually we want this to act as daemon,
 	   even when not in daemon mode.. */
@@ -571,7 +604,7 @@ main(argc, argv)
 	    if ((fd = eopen(argv[optind], O_RDWR, 0)) < 0)
 	      continue;
 	    /* the close(fd) is done in vtxprep */
-	    cfp = schedule(fd, argv[optind], tp, ino);
+	    cfp = schedule(fd, argv[optind], ino, 0);
 	    if (cfp == NULL) {
 	      if (verbose)
 		fprintf(stderr, "Nothing scheduled for %s!\n",
@@ -580,6 +613,7 @@ main(argc, argv)
 	      eunlink(argv[optind]);
 	  }
 	  doagenda();
+	  killpidfile(pidfile);
 	  exit(0);
 	}
 	mustexit = gotalarm = dumpq = rereadcf = 0;
@@ -596,7 +630,7 @@ main(argc, argv)
 	dirqueuescan(".", dirq, 1);
 
 	vtxprep_skip_lock = 0;
-	syncweb(dirq, tp);
+	syncweb(dirq);
 
 	if (dlyverbose) verbose = dlyverbose;
 
@@ -614,7 +648,7 @@ main(argc, argv)
 	  dirqueuescan(".", dirq, 1);
 	  startcount = dirq->wrksum;
 	  while (dirq->wrksum > 0 && !mustexit) {
-	    if (syncweb(dirq, tp) < 0)
+	    if (syncweb(dirq) < 0)
 	      break;
 	    queryipccheck();
 	  }
@@ -655,10 +689,10 @@ main(argc, argv)
 	  if (dirq->wrksum > 100 ) {
 	    /* If more than 100 in queue, submit 20 in a burst.. */
 	    for (i=0; i < 20; ++i)
-	      syncweb(dirq, tp);
+	      syncweb(dirq);
 	  }
 	  if (dirq->wrksum > 0)
-	    syncweb(dirq, tp);
+	    syncweb(dirq);
 
 	  /* See when to timeout from mux() */
 	  timeout = next_dirscan;
@@ -683,7 +717,7 @@ main(argc, argv)
 	      dumpq = 0;
 	    }
 	    if (rereadcf) {
-	      tp = rereadconfig(tp, config);
+	      cehead = rereadconfig(cehead, config);
 	      rereadcf = 0;
 	    }
 	    mytime(&now);
@@ -710,6 +744,9 @@ main(argc, argv)
 #ifdef	MALLOC_TRACE
 	mal_dumpleaktrace(stderr);
 #endif	/* MALLOC_TRACE */
+
+	killpidfile(pidfile);
+
 	if (mustexit)
 		die(0, "signal");
 	return 0;
@@ -800,7 +837,7 @@ dq_insert(DQ, ino, file, delay)
 	}
 
 	/* Now store the entry */
-	dsn = (struct dirstatname*)emalloc(sizeof(*dsn)+strlen(file));
+	dsn = (struct dirstatname*)emalloc(sizeof(*dsn)+strlen(file)+1);
 	memcpy(&(dsn->st),&stbuf,sizeof(stbuf));
 	dsn->ino = ino;
 	dsn->not_before = now + delay;
@@ -845,6 +882,7 @@ dq_insert(DQ, ino, file, delay)
 	  dq->wrkcount2 += 1;
 	  dq->wrksum    += 1;
 	}
+	++MIBMtaEntry->mtaReceivedMessagesSc;
 	return 0;
 }
 
@@ -986,9 +1024,8 @@ static int dirqueuescan(dir, dq, subdirs)
 	return newents;
 }
 
-static int syncweb(dq, cehead)
+static int syncweb(dq)
 	struct dirqueue *dq;
-	struct config_entry *cehead;
 {
 	struct stat *stbuf;
 	char *file;
@@ -1059,7 +1096,7 @@ static int syncweb(dq, cehead)
 	      eunlink(file);	/* hrmpf! */
 	  } else {
 	    /* Ok, schedule! */
-	    if (schedule(fd, file, cehead, ino) != NULL) {
+	    if (schedule(fd, file, ino, 0) != NULL) {
 	      /* Success, increment counters */
 	      ++wrkcnt;
 	    }
@@ -1073,13 +1110,79 @@ static int syncweb(dq, cehead)
 	return wrkcnt;
 }
 
-void resync_file(file)
+
+static int sync_cfps __((struct ctlfile *, struct ctlfile *));
+static int sync_cfps(oldcfp, newcfp)
+     struct ctlfile *oldcfp, *newcfp;
+{
+	struct vertex *ovp, *nvp;
+	struct vertex *novp, *vp;
+
+	/* Scan both files thru their   vp->next[L_CTLFILE]  vertex chains.
+	   If oldcfp has things that newcfp does not have, remove those from
+	   oldcfp chains.
+
+	   Chains start at  *cfp->head  pointer.
+
+	   Comparisons can be done in linear order with respective
+	   vp->orig[L_CHANNEL] and vp->orig[L_HOST] pointers to struct web..
+
+	   We presume the  newcfp  will not contain objects that oldcfp does
+	   not have -- while theorethically possible, it is not supported
+	   skenario..
+	*/
+
+	ovp = oldcfp->head;
+	nvp = newcfp->head;
+
+	while (ovp != NULL) {
+	  /* Always prepare for removal of the ovp object..
+	     Pick the next-ovp pointer now */
+	  novp = ovp->next[L_CTLFILE];
+
+	  /* Does this exist also on NVP chain ? */
+
+#define VTXMATCH(ovp,nvp) (((ovp)->orig[L_CHANNEL] == (nvp)->orig[L_CHANNEL]) && ((ovp)->orig[L_HOST] == (nvp)->orig[L_HOST]))
+
+	  if (!VTXMATCH(ovp,nvp)) {
+
+	    /* Uugh... Does not match :-( */
+
+	    vp = ovp;
+
+	    while (vp && !VTXMATCH(vp,nvp)) {
+	      vp = vp->next[L_CTLFILE];
+	    }
+	    if (vp == NULL) {
+	      /* New not in old at all ??? */
+	      return -1;
+	    }
+	    /* XX: All OVP instances before matching NVP
+	       are to be removed from OVP chains */
+	    /* XX: Adjust NOVP variable too */
+	  }
+	  if (VTXMATCH(ovp, nvp)) {
+	    /* XX: Verify that OVP and NVP have same amount
+	       of address indexes in them, adjust OVP if not. */
+	  }
+	  ovp = novp;
+	}
+	return 0;
+}
+
+
+void resync_file(proc, file)
+	struct procinfo *proc;
 	const char *file;
 {
 	struct stat stbuf;
 	struct spblk *spl;
 	long ino;
 	const char *s;
+	int fd;
+	struct ctlfile *oldcfp, *newcfp;
+
+	queryipccheck();
 
 	lstat(file,&stbuf);
 
@@ -1100,15 +1203,73 @@ void resync_file(file)
 	  /* Not (anymore) in processing! */
 	  return;
 	}
-	printf("Resyncing file \"%s\" (ino=%d)", file, (int) ino);
-	printf(" .. in processing db\n");
+
+	oldcfp = (struct ctlfile *)(spl->data);
+	oldcfp->id = 0; /* Don't scramble spt_mesh[] latter below */
+
+	if (spl != NULL)
+	  sp_delete(spl, spt_mesh[L_CTLFILE]);
+	spl = NULL;
+
+	printf("Resyncing file \"%s\" (ino=%d) (of=%d ho='%s')",
+	       file, (int) ino, proc->overfed,
+	       (proc->ho ? proc->ho->name : "<NULL>"));
+	/* printf(" .. in processing db\n"); */
 
 	/* cfp_free()->unvertex()->unctlfile() will do reinsertion */
 	/* dq_insert(NULL,ino,file,31); */
 
-	/* Delete it... */
-	cfp_free((struct ctlfile *)spl->data, spl);
-	/* It will return into processing at next queue scan.. */
+	/* Now read it back! */
+	fd = eopen(file, O_RDWR, 0);
+	if (fd < 0) {
+	  /* ???? */
+	  printf(" .. FILE OPEN FAILED!\n");
+
+	  /* Delete it from memory */
+	  cfp_free(oldcfp, NULL);
+	  return;
+	}
+
+	newcfp = schedule(fd, file, ino, 1);
+
+#if 0 /* XX: not usable before  sync_cfps()  works! */
+
+	if (newcfp != NULL) {
+	  /* ????  What ever, it succeeds, or it fails, all will be well */
+
+	  sync_cfps(oldcfp, newcfp);
+	  newcfp->id = 0; /* Don't scramble spt_mesh[] below */
+
+	  spl = sp_lookup((u_long)ino, spt_mesh[L_CTLFILE]);
+	  if (spl)
+	    sp_delete(spl, spt_mesh[L_CTLFILE]);
+	  oldcfp->id = ino;
+	  sp_install(oldcfp->id, (void *)oldcfp, 0, spt_mesh[L_CTLFILE]);
+
+	  printf(" .. resynced!\n");
+	} else {
+	  printf(" .. NOT resynced!\n");
+	  /* Sigh.. Throw everything away :-( */
+	  oldcfp->id = ino;
+	  cfp_free(oldcfp, NULL);
+	}
+
+	/* Delete it from memory */
+	if (newcfp != NULL)
+	  cfp_free0(newcfp);
+
+#else
+
+	if (newcfp != NULL) {
+	  printf(" .. resynced!\n");
+	} else {
+	  printf(" .. NOT resynced!\n");
+	  /* Sigh.. Throw everything away :-( */
+	  oldcfp->id = ino;
+	}
+	
+	cfp_free0(oldcfp);
+#endif
 }
 
 
@@ -1120,17 +1281,17 @@ void resync_file(file)
  * of transport directives and schedules the appropriate things to be
  * done at the appropriate time in the future.
  */
-static struct ctlfile *schedule(fd, file, cehead, ino)
+static struct ctlfile *schedule(fd, file, ino, reread)
 	int fd;
 	const char *file;
-	struct config_entry *cehead;
 	long ino;
+	const int reread;
 {
 	struct ctlfile *cfp;
 	struct vertex *vp;
 
 	/* read and process the control file */
-	if ((cfp = vtxprep(slurp(fd, ino), file)) == NULL) {
+	if ((cfp = vtxprep(slurp(fd, ino), file, reread)) == NULL) {
 	  if (!vtxprep_skip) {	/* Unless skipped.. */
 	    eunlink(file);	/* everything here has been processed */
 	    if (verbose)
@@ -1142,7 +1303,11 @@ static struct ctlfile *schedule(fd, file, cehead, ino)
 	}
 	if (cfp->head == NULL) {
 	  ++global_wrkcnt;
-	  unctlfile(cfp, 0); /* Delete the file too */
+	  ++MIBMtaEntry->mtaStoredMessages;
+	  MIBMtaEntry->mtaStoredRecipients     += cfp->rcpnts_work;
+	  MIBMtaEntry->mtaReceivedRecipientsSc += cfp->rcpnts_work;
+	  unctlfile(cfp, 0); /* Delete the file.
+				(decrements those counters too!) */
 	  return NULL;
 	}
 	for (vp = cfp->head; vp != NULL; vp = vp->next[L_CTLFILE]) {
@@ -1155,7 +1320,10 @@ static struct ctlfile *schedule(fd, file, cehead, ino)
 	}
 
 	sp_install(cfp->id, (void *)cfp, 0, spt_mesh[L_CTLFILE]);
+	++MIBMtaEntry->mtaStoredMessages;
 	++global_wrkcnt;
+	MIBMtaEntry->mtaStoredRecipients     += cfp->rcpnts_work;
+	MIBMtaEntry->mtaReceivedRecipientsSc += cfp->rcpnts_work;
 	return cfp;
 }
 
@@ -1172,7 +1340,7 @@ slurp(fd, ino)
 	register char *s;
 	register int i;
 	char *contents;
-	long *offset, *ip, *lp;
+	int *offset, *ip, *lp;
 	int offsetspace;
 	struct stat stbuf;
 	struct ctlfile *cfp;
@@ -1218,7 +1386,7 @@ slurp(fd, ino)
 	/* go through the file and mark it off */
 	i = 0;
 	offsetspace = 100;
-	offset = (long*)emalloc(sizeof(long)*offsetspace);
+	offset = (int*)emalloc(sizeof(int)*offsetspace);
 	offset[i++] = 0L;
 	for (s = contents; s - contents < stbuf.st_size; ++s) {
 	  if (*s == '\n') {
@@ -1226,7 +1394,7 @@ slurp(fd, ino)
 	    if (s - contents < stbuf.st_size) {
 	      if (i >= offsetspace-1) {
 		offsetspace += 20;
-		offset = (long*)erealloc(offset,sizeof(long)*offsetspace);
+		offset = (int*)erealloc(offset,sizeof(int)*offsetspace);
 	      }
 	      offset[i++] = s - contents;
 	      if (*s == _CF_MSGHEADERS) {
@@ -1275,13 +1443,15 @@ slurp(fd, ino)
 }
 
 struct offsort {
-	long	offset;
+	int	offset;
 	int	myidx;
-	long	headeroffset;
-	long	drptoffset;
+	int	headeroffset;
+	int	drptoffset;
+	int	delayslot;
+	int	notifyflg;
 	char	*sender;
 	/* char	*dsnrecipient; */
-	int	notifyflg;
+	time_t	wakeup;
 };
 
 /* ``bcfcn'' is used by the qsort comparison routine,
@@ -1300,9 +1470,10 @@ static int bcfcn(a, b)
 
 
 static int
-lockverify(cfp,cp)		/* Return 1 when lock process does not exist */
+lockverify(cfp,cp,verbflg)	/* Return 1 when lock process does not exist */
 	struct ctlfile *cfp;	/* Call only when the lock is marked active! */
 	const char *cp;
+	const int verbflg;
 {
 	char	lockbuf[1+_CFTAG_RCPTPIDSIZE];
 	int	lockpid;
@@ -1322,8 +1493,9 @@ lockverify(cfp,cp)		/* Return 1 when lock process does not exist */
 	if (sscanf(lockbuf,"%d",&lockpid) != 1) return 1; /* Bad value ? */
 	if (kill(lockpid,sig) != 0) return 1; /* PID does not exist, or
 					       other error.. */
-	fprintf(stderr,"lockverify: Lock with PID=%d is active on %s:%s\n",
-		lockpid, cfp->mid, cp+_CFTAG_RCPTPIDSIZE);
+	if (verbflg)
+	  fprintf(stderr,"lockverify: Lock with PID=%d is active on %s:%s\n",
+		  lockpid, cfp->mid, cp+_CFTAG_RCPTPIDSIZE);
 	return 0;	/* Locking PID does exist.. */
 }
 
@@ -1339,12 +1511,13 @@ lockverify(cfp,cp)		/* Return 1 when lock process does not exist */
  *  wrapped into same vertex node with its respective ``recipient group''.
  *
  */
-static struct ctlfile *vtxprep(cfp, file)
+static struct ctlfile *vtxprep(cfp, file, rereading)
 	struct ctlfile *cfp;
 	const char *file;
+	const int rereading;
 {
 	register int i, opcnt;
-	register long *lp;
+	register int *lp;
 	int svn;
 	char *cp, *channel, *host, *l_channel, *l_host;
 	char *echannel, *ehost, *l_echannel, *l_ehost, mfpath[100], flagstr[2];
@@ -1354,15 +1527,17 @@ static struct ctlfile *vtxprep(cfp, file)
 	struct stat stbuf;
 	struct offsort *offarr;
 	int offspc, mypid;
-	long ino;
 	int prevrcpt = -1;
 	int is_turnme = 0;
+	time_t wakeuptime;
 
 	char fpath[128], path[128], path2[128];
 
 	if (cfp == NULL)
 	  return NULL;
 
+	mytime(&now);
+
 	strcpy(fpath, file);
 
 	channel = host = NULL;
@@ -1378,13 +1553,16 @@ static struct ctlfile *vtxprep(cfp, file)
 	lp = &cfp->offset[0];
 	for (i = 0; i < cfp->nlines; ++i, ++lp) {
 	  cp = cfp->contents + *lp + 1;
+	  wakeuptime = 0;
 	  if (*cp == _CFTAG_LOCK) {
 	    /*
 	     * This can happen when we restart the scheduler, and
 	     * some previous transporter is still running.
 	     *
 	     */
-	    if (!lockverify(cfp,cp)) {
+	    if (!lockverify(cfp, cp, !rereading)) {
+#if 0
+	      long ino = 0;
 	      /*
 	       * IMO we are better off by forgetting for a while that
 	       * this spool-file exists at all.  Thus very least we
@@ -1412,20 +1590,29 @@ static struct ctlfile *vtxprep(cfp, file)
 	      dq_insert(NULL, ino, file, 32);
 
 	      return NULL;
+#else
+	      /* We can't simply forget this, we must do something
+		 smarter -- We use approach of marking the vertex
+		 non-startable until one hour from now. */
+	      wakeuptime = now + 3600; /* 1 hour */
+#endif
+	    } else {
+	      if (*cp != _CFTAG_NORMAL) {
+		*cp = _CFTAG_NORMAL; /* unlock it */
+		lockaddr(cfp->fd, NULL, (long) (cp - cfp->contents),
+			 _CFTAG_LOCK, _CFTAG_NORMAL, cfp->mid,
+			 cp+_CFTAG_RCPTPIDSIZE, mypid);
+	      }
 	    }
-	    *cp = _CFTAG_NORMAL; /* unlock it */
-	    lockaddr(cfp->fd, NULL, (long) (cp - cfp->contents),
-		     _CFTAG_LOCK, _CFTAG_NORMAL, cfp->mid,
-		     cp+_CFTAG_RCPTPIDSIZE, mypid);
 	  }
 	  /* Calculate summary info */
 	  if (cp[-1] == _CF_RECIPIENT) {
-	    cfp->rcpnts_total += 1;
+	    ++cfp->rcpnts_total;
 	    if (*cp == _CFTAG_NOTOK) {
-	      cfp->rcpnts_failed += 1;
+	      ++cfp->rcpnts_failed;
 	      prevrcpt = -1;
-	    } else if (*cp == _CFTAG_OK) {
-	      cfp->rcpnts_work   += 1;
+	    } else if (*cp != _CFTAG_OK) {
+	      ++cfp->rcpnts_work;
 	    }
 	  }
 	  if (*cp == _CFTAG_NORMAL ||
@@ -1464,7 +1651,16 @@ static struct ctlfile *vtxprep(cfp, file)
 	      if (*cp == ' ' || (*cp >= '0' && *cp <= '9')) {
 		/* New PID locking scheme.. */
 		offarr[opcnt].offset += _CFTAG_RCPTPIDSIZE;
+		cp += _CFTAG_RCPTPIDSIZE;
 	      }
+	      if (*cp == ' ' || (*cp >= '0' && *cp <= '9')) {
+		/* Newer DELAY data slot - _CFTAG_RCPTDELAYSIZE bytes */
+		offarr[opcnt].delayslot = offarr[opcnt].offset;
+		offarr[opcnt].offset += _CFTAG_RCPTDELAYSIZE;
+		cp += _CFTAG_RCPTDELAYSIZE;
+	      } else
+		offarr[opcnt].delayslot = 0;
+	      offarr[opcnt].wakeup = wakeuptime;
 	      offarr[opcnt].myidx = i;
 	      offarr[opcnt].headeroffset = -1;
 	      offarr[opcnt].drptoffset = -1;
@@ -1565,10 +1761,6 @@ static struct ctlfile *vtxprep(cfp, file)
 	}
 	close(cfp->fd);	/* closes the fd opened in schedule() */
 
-	if (is_turnme && cfp->mid == NULL)
-	  unlink(file);
-
-
 	if (fpath[0] >= 'A') {
 	  /* Prefixed with a subdir path */
 	  int hash = 0;
@@ -1643,9 +1835,9 @@ static struct ctlfile *vtxprep(cfp, file)
 	  else
 	    sprintf(mfpath, "../%s/%s", QUEUEDIR, cfp->mid);
 	}
-	if (cfp->mid == NULL || cfp->logident == NULL
-	    || estat(mfpath, &stbuf) < 0) {
-	  if (cfp->vfpfn != NULL) {
+	if (cfp->mid == NULL || cfp->logident == NULL ||
+	    estat(mfpath, &stbuf) < 0) {
+	  if (!is_turnme && cfp->vfpfn != NULL) {
 	    FILE *vfp = vfp_open(cfp);
 	    if (vfp) {
 	      fprintf(vfp,
@@ -1657,7 +1849,7 @@ static struct ctlfile *vtxprep(cfp, file)
 	  free(offarr);
 	  return NULL;
 	}
-	cfp->ctime = stbuf.st_mtime; /* instead of ctime, we use mtime
+	cfp->mtime = stbuf.st_mtime; /* instead of ctime, we use mtime
 					at the queue-dir accesses, this
 					way we can move the spool to
 					another machine, and run things
@@ -1719,6 +1911,7 @@ static struct ctlfile *vtxprep(cfp, file)
 #endif
 	    {
 	      ehost = skip821address(host);
+	      strlower(host);
 	      if (ehost == NULL || ehost == host) /* error! */
 		continue;
 	    }
@@ -1740,13 +1933,13 @@ static struct ctlfile *vtxprep(cfp, file)
 	      vp->retryindex   = 0;
 	      vp->nextitem     = NULL;
 	      vp->previtem     = NULL;
-	      vp->wakeup       = 0;
 	      vp->proc         = NULL;
 	      vp->attempts     = 0;
 	      vp->notary       = NULL;
 #endif
 	      vp->ngroup       = i - svn;
 	      /* vp->sender       = strsave(offarr[svn].sender); */
+	      vp->wakeup       = offarr[svn].wakeup;
 	      vp->headeroffset = offarr[svn].headeroffset; /*They are similar*/
 	      vp->drptoffset   = offarr[svn].drptoffset;
 	      vp->notaryflg    = offarr[svn].notifyflg;
@@ -1787,12 +1980,12 @@ static struct ctlfile *vtxprep(cfp, file)
 	  vp->nextitem = NULL;
 	  vp->previtem = NULL;
 	  vp->proc = NULL;
-	  vp->wakeup = 0;
 	  vp->attempts     = 0;
 	  vp->notary       = NULL;
 #endif
 	  vp->ngroup = i - svn;
 	  /* vp->sender = strsave(offarr[snv].sender); */
+	  vp->wakeup       = offarr[svn].wakeup;
 	  vp->headeroffset = offarr[svn].headeroffset; /* Just any of them will do */
 	  vp->drptoffset = offarr[svn].drptoffset;
 	  vp->notaryflg  = offarr[svn].notifyflg;
@@ -1909,9 +2102,9 @@ static void ce_fillin(thg,cep)
  *
  */
 
-static void vtxdo(vp, cehead, path)
+static void vtxdo(vp, cehdr, path)
 	struct vertex *vp;
-	struct config_entry *cehead;
+	struct config_entry *cehdr;
 	const char *path;
 {
 	struct config_entry *tp;
@@ -1923,7 +2116,7 @@ static void vtxdo(vp, cehead, path)
 	 * fill in the blanks in the vertex specification
 	 */
 	n = 0;
-	for (tp = cehead; tp != NULL; tp = tp->next) {
+	for (tp = cehdr; tp != NULL; tp = tp->next) {
 	  ++cnt;
 	  if (vtxmatch(vp, tp)) {
 	    /* tp points to the selected config entry */
@@ -1943,7 +2136,7 @@ static void vtxdo(vp, cehead, path)
 
 	/* set default values */
 	if (tp->expiry > 0)
-	  vp->ce_expiry = tp->expiry + vp->cfp->ctime;
+	  vp->ce_expiry = tp->expiry + vp->cfp->mtime;
 	else
 	  vp->ce_expiry = 0;
 
@@ -2032,7 +2225,7 @@ static int globmatch(pattern, string)
 		int len = strlen(string);
 		i = strlen(pattern);
 		if (i > len) return 0; /* Tough.. pattern longer than string */
-		if (strcmp(string +(len-i),pattern) == 0)
+		if (memcmp(string + len - i, pattern, i) == 0)
 		  return 1; /* MATCH! */
 	      }
 	    }
