@@ -34,9 +34,21 @@
 #endif
 #include <sys/types.h>
 #include <sys/stat.h>
+#ifdef HAVE_STRING_H
+#include <string.h>
+#else
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
+#endif
+
+#include "zmalloc.h"
+#include "libz.h"
 #include "ta.h"
 
+#ifndef strdup
 extern char *strdup();
+#endif
 #ifndef strchr
 extern char *strchr();
 #endif
@@ -157,12 +169,12 @@ struct rcpt *rp;
 #ifdef HAVE_STDARG_H
 #ifdef __STDC__
 int
-append_header(struct rcpt *rp, char *fmt, ...)
+append_header(struct rcpt *rp, const char *fmt, ...)
 #else /* Not ANSI-C */
 int
 append_header(rp, fmt)
 	struct rcpt *rp;
-	char *fmt;
+	const char *fmt;
 #endif
 #else
 int
@@ -171,10 +183,10 @@ append_header(va_alist)
 #endif
 {
 	va_list pvar;
-	char linebuf[2000]; /* XX: SHOULD be enough..  damn vsprintf()..*/
-	int linelen;
-	int linecnt;
-	char **hdr, ***hdrpp, **hdrp2;
+	char lbuf[2000]; /* XX: SHOULD be enough..  damn vsprintf()..*/
+	u_int linelen;
+	u_int linecnt;
+	char ***hdrpp, **hdrp2;
 
 #ifdef HAVE_STDARG_H
 	va_start(pvar,fmt);
@@ -186,17 +198,17 @@ append_header(va_alist)
 	rp  = va_arg(pvar, struct rcpt*);
 	fmt = va_arg(pvar, char *);
 #endif
-	linebuf[0] = 0;
+	lbuf[0] = 0;
 
 #ifdef HAVE_VSNPRINTF
-	vsnprintf(linebuf, sizeof(linebuf)-1, fmt, pvar);
+	vsnprintf(lbuf, sizeof(lbuf)-1, fmt, pvar);
 #else
-	vsprintf(linebuf, fmt, pvar);
+	vsprintf(lbuf, fmt, pvar);
 #endif
 	va_end(pvar);
 
-	linelen = strlen(linebuf);
-	if (linelen > sizeof(linebuf)) {
+	linelen = strlen(lbuf);
+	if (linelen > sizeof(lbuf)) {
 	  exit(240); /* BUG TIME! */
 	}
 
@@ -216,7 +228,7 @@ append_header(va_alist)
 
 	if (!hdrp2) return -1;
 	hdrp2[linecnt] = (char*) emalloc(linelen+3);
-	memcpy(hdrp2[linecnt],linebuf,linelen+2);
+	memcpy(hdrp2[linecnt],lbuf,linelen+2);
 	hdrp2[++linecnt] = NULL;
 	*hdrpp = hdrp2;
 	return linecnt;
@@ -364,7 +376,7 @@ struct ct_data *
 parse_content_type(ct_linep)
 char **ct_linep;	/* Could be multiline! */
 {
-	char *line, *s, *p, *pv, *ss;
+	char *s, *p, *pv, *ss;
 	struct ct_data *ct = (struct ct_data*)emalloc(sizeof(struct ct_data));
 	int unknowncount = 0;
 
@@ -534,7 +546,6 @@ char **ct_linep;	/* Could be multiline! */
 	  }
 	}
 	/*NOTREACHABLE*/
-	abort();
 }
 
 struct cte_data *
@@ -593,8 +604,8 @@ struct rcpt *rp;
 	return 0;	/* No "Content-Conversion:" header */
 }
 
-static char *cCTE = "Content-Transfer-Encoding:";
-static char *cCT  = "Content-Type:";
+static const char *cCTE = "Content-Transfer-Encoding:";
+static const char *cCT  = "Content-Type:";
 
 int
 cte_check(rp)
@@ -632,7 +643,7 @@ struct rcpt *rp;
 char **  /* Return a pointer to header line pointer */
 has_header(rp,keystr)
 struct rcpt *rp;
-char *keystr;
+const char *keystr;
 {
 	char **hdrs = *(rp->newmsgheader);
 	int keylen = strlen(keystr);
@@ -705,28 +716,27 @@ FILE *verboselog;
 	return 1;
 }
 
-void
+int
 downgrade_headers(rp, convertmode, verboselog)
 struct rcpt *rp;
 int convertmode;
 FILE *verboselog;
 {
-	char *buf, *nbuf, *bufo;
 	char ***oldmsgheader;
 	char **CT   = NULL;
 	char **CTE  = NULL;
 	char **MIME = NULL;
-	char **receivedp = NULL;
+	/* char **receivedp = NULL; */
 	struct ct_data *ct;
 	int lines = 0;
 	int i;
+	int is_textplain;
 	int newlen;
-	int ct_offset = 0;
 
 	if (*(rp->newmsgheadercvt) != NULL)
-	  return; /* Already converted ! */
+	  return 0; /* Already converted ! */
 
-	if (!cvtspace_copy(rp)) return; /* XX: auch! */
+	if (!cvtspace_copy(rp)) return 0; /* XX: auch! */
 
 	oldmsgheader = rp->newmsgheadercvt;
 
@@ -742,7 +752,7 @@ FILE *verboselog;
 
 	if (convertmode == _CONVERT_UNKNOWN) {
 	  /* We downgrade by changing it to Q-P as per RFC 1428/Appendix A */
-	  static char *warning_lines[] = {
+	  static const char *warning_lines[] = {
 "X-Warning: Original message contained 8-bit characters, however during",
 "           the SMTP transport session the receiving system was unable to",
 "           announce capability of receiving 8-bit SMTP (RFC 1651-1653),",
@@ -769,7 +779,7 @@ NULL };
 #endif
 	  sprintf(newct,"Content-Type: TEXT/PLAIN; charset=%s",defcharset);
 
-	  if (!newmsgheaders) return; /* XX: Auch! */
+	  if (!newmsgheaders) return 0; /* XX: Auch! */
 
 	  for (lines = 0; warning_lines[lines] != NULL; ++lines)
 	    newmsgheaders[lines] = strdup(warning_lines[lines]);
@@ -797,14 +807,14 @@ NULL };
 
 	  free(*oldmsgheader); /* Free the old one.. */
 	  *oldmsgheader = newmsgheaders;
-	  return;
+	  return 0;
 	}
 
 	/* Now look for the  Content-Transfer-Encoding:  header */
 
-	receivedp = has_header(rp,"Received:");
+	/*receivedp = has_header(rp,"Received:");*/
 
-	if (CTE == NULL) return; /* No C-T-E: ??? */
+	if (CTE == NULL) return 0; /* No C-T-E: ??? */
 
 	/* strlen("Content-Transfer-Encoding: QUOTED-PRINTABLE") == 43
 	   strlen("Content-Transfer-Encoding: 7BIT") == 31		*/
@@ -815,44 +825,52 @@ NULL };
 
 	*CTE = (char *)ctlrealloc(rp->desc,*CTE,newlen+2);
 
-	if (convertmode == _CONVERT_QP) {
-
-	  strcpy(*CTE, "Content-Transfer-Encoding: QUOTED-PRINTABLE");
-	  mime_received_convert(rp," convert rfc822-to-quoted-printable");
-	  return; /* No change on   Charset.. */
-
-	} else
-	  strcpy(*CTE, "Content-Transfer-Encoding: 7BIT");
-
 	/* Ok, this was C-T-E: 7BIT, turn charset to US-ASCII if it
 	   was  ISO-*  */
 	if (CT == NULL && verboselog) {
 	  fprintf(verboselog,"Had Content-Transfer-Encoding -header, but no Content-Type header ???  Adding C-T..\n");
 	}
-	
+
 	if (CT == NULL) { /* ???? Had C-T-E, but no C-T ?? */
 	  append_header(rp,"Content-Type: TEXT/PLAIN; charset=US-ASCII");
-	  return;
+	  return 0;
 	}
 
 	ct = parse_content_type(CT);
 
-	if (ct->basetype == NULL ||
-	    ct->subtype  == NULL ||
-	    cistrcmp(ct->basetype,"text") != 0 ||
-	    cistrcmp(ct->subtype,"plain") != 0) return; /* Not TEXT/PLAIN! */
+	is_textplain = (ct->basetype != NULL &&
+			ct->subtype  != NULL &&
+			cistrcmp(ct->basetype,"text") == 0 &&
+			cistrcmp(ct->subtype,"plain") == 0);
 
-	if (ct->charset &&
-	    cistrncmp(ct->charset,"ISO-8859",8) != 0 &&
-	    cistrncmp(ct->charset,"KOI8",4)     != 0) return; /* Not ISO-* */
+	if (ct->charset && is_textplain &&
+	    (convertmode != _CONVERT_QP) &&
+	    /* Change to US-ASCII for known 7-bit clean
+	       inputs where the claimed charset(prefix) has
+	       all its charsets equal to US-ASCII in the
+	       low 128 characters. */
+	    (cistrncmp(ct->charset,"ISO-8859",8) == 0 ||
+	     cistrncmp(ct->charset,"KOI8",4)     == 0)) {
 
-	if (ct->charset)
-	  free(ct->charset);
+	  if (ct->charset)
+	    free(ct->charset);
 
-	ct->charset = strdup("US-ASCII");
+	  ct->charset = strdup("US-ASCII");
+	  strcpy(*CTE, "Content-Transfer-Encoding: 7BIT");
+
+	}
 
 	/* Delete the old one, and place there the new version.. */
 	output_content_type(rp,ct,CT);
+
+	if (convertmode == _CONVERT_QP) {
+
+	  strcpy(*CTE, "Content-Transfer-Encoding: QUOTED-PRINTABLE");
+	  mime_received_convert(rp," convert rfc822-to-quoted-printable");
+
+	}
+
+	return 1; /* Non-zero for success! */
 }
 
 
@@ -906,7 +924,8 @@ qp_to_8bit(rp)
 {
 	char **inhdr;
 	char *hdr, *p;
-	char **CTE;
+	char **CTE, **CT;
+	struct ct_data *ct;
 
 	if (!cvtspace_copy(rp))
 	  return 0;	/* Failed to copy ! */
@@ -914,8 +933,16 @@ qp_to_8bit(rp)
 	inhdr = *(rp->newmsgheadercvt);
 
 	CTE = has_header(rp,cCTE);
+	CT  = has_header(rp,cCT);
 
-	if (!CTE) return 0; /* No C-T-E header! */
+	if (!CTE || !CT) return 0; /* No C-T-E header! */
+
+	ct = parse_content_type(CT);
+
+	if (ct->basetype == NULL ||
+	    ct->subtype  == NULL ||
+	    cistrcmp(ct->basetype,"text") != 0 ||
+	    cistrcmp(ct->subtype,"plain") != 0) return 0; /* Not TEXT/PLAIN! */
 
 	hdr = *CTE;
 
