@@ -26,7 +26,11 @@
 #undef datum
 #endif
 #ifdef HAVE_DB_H
-#include <db.h>
+#if defined(HAVE_DB_185_H) && !defined(HAVE_DB_OPEN2)
+# include <db_185.h>
+#else
+# include <db.h>
+#endif
 #endif
 
 #define PROG "makedb"
@@ -40,6 +44,9 @@ char *progname    = NULL;
 int   D_alloc     = 0;
 int   verbose     = 0;
 int   append_mode = 0;
+int   lc_key	  = 0;
+int   uc_key      = 0;
+int   silent      = 0;
 
 /* extern char *strchr(); */
 
@@ -51,7 +58,7 @@ void usage(prog, errs, err)
 const char *prog, *errs;
 int err;
 {
-	fprintf(stderr, "Usage: %s [-A][-a|-p] dbtype database.name [infilename|-]\n", prog);
+	fprintf(stderr, "Usage: %s [-l|-u]] [-A][-a|-p][-s] dbtype database.name [infilename|-]\n", prog);
 	fprintf(stderr, "  where supported dbtypes are:");
 #ifdef HAVE_NDBM_H
 	fprintf(stderr, " ndbm");
@@ -63,7 +70,13 @@ int err;
 	fprintf(stderr, " btree bhash");
 #endif
 	fprintf(stderr, "\n");
-	fprintf(stderr, "  If no infilename is defined, database.name is assumed.\n");
+	fprintf(stderr, " Error now: %s", errs);
+	if (err != 0) {
+	  fprintf(stderr, ", errno=%d (%s)", err, strerror(err));
+	  fprintf(stderr, "\n");
+	  exit(1);
+	}
+	fprintf(stderr, "\n\n  If no infilename is defined, database.name is assumed.\n");
 #ifdef HAVE_NDBM_H
 	fprintf(stderr, "  (NDBM appends  .pag, and .dir  into actual db file names..)\n");
 #endif
@@ -75,15 +88,18 @@ int err;
 	fprintf(stderr, "  (BHASH appends .pag, and .dir into actual db file names..)\n");
 #endif
 	fprintf(stderr, "\n");
-	fprintf(stderr, "  The '-a' option is for parsing input that comes in\n");
-	fprintf(stderr, "  'aliases' format:  'key: data,in,long,line,'\n");
-	fprintf(stderr, "                     '  with,indented,extended,lines'\n");
-	fprintf(stderr, "  The '-p' option is for parsing smtpserver policy database.\n");
-	fprintf(stderr, "  The '-A' option APPENDS new data to existing keyed data.\n");
-	fprintf(stderr, " Error now: %s", errs);
-	fprintf(stderr, ", errno=%d (%s)", err, strerror(err));
-	fprintf(stderr, "\n");
-	exit(1);
+
+	fprintf(stderr,
+"  The '-a' option is for parsing input that comes in\n\
+  'aliases' format:  'key: data,in,long,line,'\n\
+                     '  with,indented,extended,lines'\n\
+  The '-p' option is for parsing smtpserver policy database.\n\
+  The '-A' option APPENDS new data to existing keyed data.\n\
+  The '-l' and '-u' will lower-/uppercasify key string before\n\
+    storing it into the database.  This does not apply to '-p'.\n\
+  The '-s' option orders 'silent running' -- report only errors.\n");
+
+	exit(64);
 }
 
 
@@ -107,7 +123,8 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
 	int rc = -2;
 
 	if (verbose)
-	  fprintf(stderr,"Storing key='%s' data='%s'\n", t, s);
+	  fprintf(stderr, "Storing key='%s' data='%s'\n",
+		  (const char*)t, (const char *)s);
 
 #ifdef HAVE_NDBM_H
 	if (typ == 1) {
@@ -134,12 +151,23 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
 #ifdef HAVE_DB_H
 	if (typ == 3 || typ == 4) {
 		DBT Bkey, Bdat;
+
+		memset(&Bkey,0,sizeof(Bkey));
+		memset(&Bdat,0,sizeof(Bdat));
+
 		Bkey.data = (void*)t;
 		Bkey.size = tlen;
+
 		Bdat.data = (void*)s;
 		Bdat.size = slen;
+
+#ifdef HAVE_DB_OPEN2
+		rc = (dbfile->put) (dbfile, NULL, &Bkey, &Bdat,
+				    overwritemode ? 0: DB_NOOVERWRITE);
+#else
 		rc = (dbfile->put) (dbfile, &Bkey, &Bdat,
 				    overwritemode ? 0: R_NOOVERWRITE);
+#endif
 	}
 #endif
 
@@ -178,9 +206,17 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
 #ifdef HAVE_DB_H
 	  if (typ == 3 || typ == 4) {
 	    DBT Bkey, Bdat;
+
+	    memset(&Bkey,0,sizeof(Bkey));
+	    memset(&Bdat,0,sizeof(Bdat));
+
 	    Bkey.data = (void*)t;
 	    Bkey.size = tlen;
+#ifdef HAVE_DB_OPEN2
+	    rc = (dbfile->get) (dbfile, NULL, &Bkey, &Bdat, 0);
+#else
 	    rc = (dbfile->get) (dbfile, &Bkey, &Bdat, 0);
+#endif
 	    if (rc != 0)
 	      memset(&Bdat, 0, sizeof(Bdat));
 	    dataptr = Bdat.data;
@@ -272,7 +308,7 @@ char **sout;
 	return start;
 }
 
-void create_policy_dbase(infile, dbf, typ)
+int create_policy_dbase(infile, dbf, typ)
 FILE *infile;
 void *dbf;
 const int typ;
@@ -281,6 +317,7 @@ const int typ;
 	char policydata[16*1024];
 	char *s, *t, *str1, *str2;
 	int rc, tlen, slen, llen;
+	int errflag = 0;
 
 	int policydatalen = 0;
 	int linenum = 0;
@@ -289,28 +326,29 @@ const int typ;
 	   store_db() to work differently from alias append */
 	append_mode = -append_mode;
 
-	while ((llen = getline(infile)) != 0) {
+	while ((llen = zgetline(infile)) != 0) {
 		++linenum;
 		policydatalen = 0;
-		if (linebuf[llen-1] != '\n') {
+		if (zlinebuf[llen-1] != '\n') {
 			/* Eh ? No line ending newline ?
 			   Ah well, there is always at least one byte
 			   of space after the read block. */
 			++llen;
 			fprintf(stderr, "input line of len %d lacking trailing newline, corrupted file ?\n", llen-1);
+			errflag = 1;
 		}
-		linebuf[llen-1] = '\0';
+		zlinebuf[llen-1] = '\0';
 
-		if (*linebuf == '#')
+		if (*zlinebuf == '#')
 			continue;	/* Comment! */
 
-		if (*linebuf == ' ' || *linebuf == '\t')
+		if (*zlinebuf == ' ' || *zlinebuf == '\t')
 			continue;	/* Starts with a white-space! */
 
 		/* Scan first white-space separated token,
 		   point its start with t! */
 
-		t = linebuf;
+		t = zlinebuf;
 		while (*t == '\t' || *t == ' ')
 			++t;
 
@@ -335,6 +373,7 @@ const int typ;
 			fprintf(stderr,
 				"Error: line %d: bad policykey, rc = %d\n",
 				linenum, rc);
+			errflag = 1;
 			continue;
 		}
 		/* Skip LWSP */
@@ -348,6 +387,7 @@ const int typ;
 			fprintf(stderr,
 				"Error: No attribute pair on line %d.\n",
 				linenum);
+			errflag = 1;
 			continue;
 		} else {
 			int err = 0;
@@ -355,16 +395,19 @@ const int typ;
 				str2 = tokskip(s, &s);
 				if (str2 == NULL) {
 					fprintf(stderr,
-						"Error: Invalid attribute pair on line %d.\n",
-						linenum);
+						"Error: Invalid attribute pair on line %d; aname='%s', value missing.\n",
+						linenum, str1);
+					errflag = 1;
+					err = 1;
 					break;
 				}
 				rc = parseattributepair((void *) &policydata[policydatalen],
 							str1, str2);
 				if (rc != 0) {
 					fprintf(stderr,
-						"Error: Invalid attribute pair on line %d.\n",
-						linenum);
+						"Error: Invalid attribute pair (\"%s %s\")  on line %d.\n",
+						str1, str2, linenum);
+					errflag = 1;
 					err = 1;
 					break;
 				}
@@ -388,19 +431,19 @@ const int typ;
 			      t, tlen, s, slen);
 		if (rc > 0) {
 			int tl = tlen;
-			fprintf(stderr, "Error: Duplicate key at line %d: \"",
+			fprintf(stderr, "WARNING: Duplicate key at line %d: \"",
 				linenum);
-			for (;tlen > 0; --tlen,++t) {
+			for (;tl > 0; --tl,++t) {
 				unsigned char c = *t;
 				if (c < ' ' || c > 126 || c == '\\')
-					fprintf(stderr, "\0%03o", c);
+					fprintf(stderr, "\\%03o", c);
 				else
 					fprintf(stderr, "%c", c);
 			}
 			fprintf(stderr, "\"\n");
 		}
 	}
-	return;
+	return errflag;
 }
     
 
@@ -443,7 +486,7 @@ skipaliastoken(s)
 }
 
 
-void create_aliases_dbase(infile, dbf, typ)
+int create_aliases_dbase(infile, dbf, typ)
 FILE *infile;
 void *dbf;
 const int typ;
@@ -451,31 +494,45 @@ const int typ;
 	int tlen, slen, llen;
 	char *s, *t;
 	char *t0 = NULL, *s0 = NULL;
-	int linenum = 0;
+	int  linenum = 0;
+	int  errflag = 0;
+	int  longest = 0;
+	int  count   = 0;
+	long totsize = 0;
 
-	while ((llen = getline(infile)) != 0) {
+	while ((llen = zgetline(infile)) != 0) {
 		++linenum;
-		if (linebuf[llen-1] != '\n') {
+		if (zlinebuf[llen-1] != '\n') {
 			/* Eh ? No line ending newline ?
 			   Ah well, there is always at least one byte
 			   of space after the read block. */
 			++llen;
 			fprintf(stderr, "input line lacking trailing newline, corrupted file ?\n");
-			fprintf(stderr, " len=%d, linebuf='%s'\n", llen, linebuf);
+			fprintf(stderr, " len=%d, zlinebuf='%s'\n", llen, zlinebuf);
 		}
-		linebuf[llen-1] = '\0';
+		zlinebuf[llen-1] = '\0';
 
 		if (verbose)
-		  fprintf(stderr,"aliases parser: getline() llen=%d, str='%.*s'\n",
-			  llen, llen, linebuf);
+		  fprintf(stderr,"aliases parser: zgetline() llen=%d, str='%.*s'\n",
+			  llen, llen, zlinebuf);
 
-		if (*linebuf == '#') {
+		if (*zlinebuf == '#') {
 			/* Comment! */
 		store_and_continue:
 			if (t0 != NULL) {
 				tlen = strlen(t0) + 1;
 				slen = strlen(s0) + 1;
 
+				++count;
+				totsize += slen;
+				if (longest < slen)
+				  longest = slen;
+
+				if (uc_key)
+				  strupper(t0); /* Uppercasify the key */
+				if (lc_key)
+				  strlower(t0); /* Lowercasify the key */
+
 				if (store_db(dbf, typ, 0, linenum,
 					     t0, tlen, s0, slen) < 0)
 					break;
@@ -485,7 +542,7 @@ const int typ;
 			continue;
 		}
 
-		t = linebuf;
+		t = zlinebuf;
 		/* Key starts at line start, continuation lines start
 		   with white-space */
 
@@ -496,15 +553,28 @@ const int typ;
 			/* Continuation line */
 			while (*t == '\t' || *t == ' ') ++t;
 			slen = strlen(s0);
-			tlen = strlen(t) + 2;
-			s0   = erealloc(s0, slen + tlen + 4);
-			memcpy(s0 + slen + 0, " ,\n\t", 4);
-			memcpy(s0 + slen + 4, t, tlen-1);
+			tlen = strlen(t);
+			for (llen = slen; llen > 0; --llen) {
+			  /* Chop trailing white-space, if any */
+			  if (s0[llen-1] == ' ' || s0[llen-1] == '\t') {
+			    s0[llen-1] = '\0';
+			  } else
+			    break;
+			}
+			slen = llen; /* Shortened, possibly.. */
+			if (s0[llen-1] != ',') {
+			  fprintf(stderr, "Line %d: Continuation line on alias without preceeding line ending with comma (',')\n", linenum);
+			  errflag = 1;
+			  continue;
+			}
+			s0   = erealloc(s0, slen + tlen + 1);
+			memcpy(s0 + slen, t, tlen + 1); /* end NIL included */
 			continue;
 		}
 		if (*t == '\t' || *t == ' ') {
-			/* Continuaton line without previous key line */
+			/* Continuation line without previous key line */
 			fprintf(stderr,"Line %d: Continuation line without initial keying line\n", linenum);
+			errflag = 1;
 			continue;
 		}
 
@@ -514,6 +584,16 @@ const int typ;
 			tlen = strlen(t0) + 1;
 			slen = strlen(s0) + 1;
 
+			++count;
+			totsize += slen;
+			if (longest < slen)
+			  longest = slen;
+
+			if (uc_key)
+			  strupper(t0); /* Uppercasify the key */
+			if (lc_key)
+			  strlower(t0); /* Lowercasify the key */
+
 			if (store_db(dbf, typ, 0, linenum,
 				     t0, tlen, s0, slen) < 0)
 				break;
@@ -530,6 +610,7 @@ const int typ;
 			/* This alias-token is invalid */
 			fprintf(stderr,"Line %d: Invalid alias key token; missing colon ?\n",
 				linenum);
+			errflag = 1;
 			continue;
 		}
 
@@ -548,15 +629,28 @@ const int typ;
 		tlen = strlen(t0) + 1;
 		slen = strlen(s0) + 1;
 
+		++count;
+		totsize += slen;
+		if (longest < slen)
+		  longest = slen;
+
+		if (uc_key)
+		  strupper(t0); /* Uppercasify the key */
+		if (lc_key)
+		  strlower(t0); /* Lowercasify the key */
+
 		(void) store_db(dbf, typ, 0, linenum,
 				t0, tlen, s0, slen);
 		if (t0) free(t0);  t0 = NULL;
 		if (s0) free(s0);  s0 = NULL;
 	}
-	return;
+	if (!silent)
+	  fprintf(stdout, "%d aliases, longest %d bytes, %ld bytes total\n",
+		  count, longest, totsize);
+	return errflag;
 }
 
-void create_keyed_dbase(infile, dbf, typ)
+int create_keyed_dbase(infile, dbf, typ)
 FILE *infile;
 void *dbf;
 const int typ;
@@ -564,25 +658,26 @@ const int typ;
 	int tlen, slen, llen;
 	char *s, *t;
 	int linenum = 0;
+	int errflag = 0;
 
-	while ((llen = getline(infile)) != 0) {
+	while ((llen = zgetline(infile)) != 0) {
 		++linenum;
-		if (linebuf[llen-1] != '\n') {
+		if (zlinebuf[llen-1] != '\n') {
 			/* Eh ? No line ending newline ?
 			   Ah well, there is always at least one byte
 			   of space after the read block. */
 			++llen;
 			fprintf(stderr, "input file lacking trailing newline, corrupted file ?\n");
 		}
-		linebuf[llen-1] = '\0';
+		zlinebuf[llen-1] = '\0';
 
-		if (*linebuf == '#')
+		if (*zlinebuf == '#')
 			continue;	/* Comment! */
 
 		/* Scan first white-space separated token,
 		   point its start with t! */
 
-		t = linebuf;
+		t = zlinebuf;
 		while (*t == '\t' || *t == ' ')
 			++t;
 
@@ -604,15 +699,20 @@ const int typ;
 		while (*s && (*s == '\t' || *s == ' '))
 			++s;
 	    
-		strlower(t); /* Lowercasify the key */
+		if (uc_key)
+		  strupper(t); /* Uppercasify the key */
+		if (lc_key)
+		  strlower(t); /* Lowercasify the key */
 	    
 		tlen = strlen(t) + 1;
 		slen = strlen(s) + 1;
 		if (store_db(dbf, typ, 0, linenum,
-			     t, tlen, s, slen) < 0)
-			break;
+			     t, tlen, s, slen) < 0) {
+		  errflag = 1; /* Duplicate ?? */
+		  break;
+		}
 	}
-	return;
+	return errflag;
 }
 
 
@@ -622,7 +722,7 @@ char *argv[];
 {
     char *dbasename = NULL;
     FILE *infile = NULL;
-    int c;
+    int c, rc;
     int typ = 0;
 #ifdef HAVE_NDBM_H
     DBM *ndbmfile = NULL;
@@ -642,8 +742,14 @@ char *argv[];
 
     progname = argv[0];
 
-    while ((c = getopt(argc, argv, "Aapv")) != EOF) {
+    while ((c = getopt(argc, argv, "Aalpsuv")) != EOF) {
 	switch (c) {
+	case 'l':
+	    lc_key = 1;
+	    break;
+	case 'u':
+	    uc_key = 1;
+	    break;
 	case 'A':
 	    append_mode = 1;
 	    break;
@@ -653,6 +759,9 @@ char *argv[];
 	case 'p':
 	    policyinput = 1;
 	    break;
+	case 's':
+	    silent = 1;
+	    break;
 	case 'v':
 	    verbose = 1;
 	    break;
@@ -722,6 +831,22 @@ char *argv[];
     }
 #endif
 #ifdef HAVE_DB_H
+#ifdef HAVE_DB_OPEN2
+    if (typ == 3) {
+	dbasename = strcpy(malloc(strlen(dbasename) + 8), dbasename);
+	strcat(dbasename, ".db");	/* ALWAYS append this */
+	dbfile = NULL;
+	db_open(dbasename, DB_BTREE,  DB_CREATE|DB_TRUNCATE,
+		0644, NULL, NULL, &dbfile);
+	dbf = dbfile;
+    }
+    if (typ == 4) {
+	dbfile = NULL;
+	db_open(dbasename, DB_HASH,  DB_CREATE|DB_TRUNCATE,
+		0644, NULL, NULL, &dbfile);
+	dbf = dbfile;
+    }
+#else
     if (typ == 3) {
 	dbasename = strcpy(malloc(strlen(dbasename) + 8), dbasename);
 	strcat(dbasename, ".db");	/* ALWAYS append this */
@@ -734,18 +859,19 @@ char *argv[];
 			DB_HASH, NULL);
 	dbf = dbfile;
     }
+#endif
 #endif
     if (dbf == NULL)
 	usage(argv0, "Can't open dbase file", errno);
 
-    initline(BUFSIZ);
+    initzline(BUFSIZ);
 
     if (policyinput)
-	    create_policy_dbase(infile, dbf, typ);
+	    rc = create_policy_dbase(infile, dbf, typ);
     else if (aliasinput)
-	    create_aliases_dbase(infile, dbf, typ);
+	    rc = create_aliases_dbase(infile, dbf, typ);
     else
-	    create_keyed_dbase(infile, dbf, typ);
+	    rc = create_keyed_dbase(infile, dbf, typ);
 
 #ifdef HAVE_NDBM_H
     if (typ == 1)
@@ -758,9 +884,13 @@ char *argv[];
 #ifdef HAVE_DB_H
     if (typ == 3 || typ == 4) {
 	(dbfile->sync) (dbfile, 0);
+#ifdef HAVE_DB_OPEN2
+	(dbfile->close) (dbfile, 0);
+#else
 	(dbfile->close) (dbfile);
+#endif
     }
 #endif
 
-    return 0;
+    return (rc ? 1 : 0);
 }
